#############################################################################
#############################################################################
#
# latexEngine.tcl (called from latex.tcl)
#
# The LaTeX engine
#
#############################################################################
#
# Old Author:  Tom Scavo <trscavo@syr.edu>
# New Author:  Vince Darley <darley@fas.harvard.edu>
#
#############################################################################
#############################################################################
proc latexEngine.tcl {} {}


#--------------------------------------------------------------------------
# Mark Menu:
#--------------------------------------------------------------------------

# Bugs:  -- won't add a section (in a different chapter) with the same name.
#        -- should ignore comments, but doesn't.
#        -- a sectioning command must be on a line by itself.
proc TeX::MarkFile {} {
	if {[file extension [win::CurrentTail]] == ".sty"} {
		return [TeX::Sty_MarkFile]
	}
	set pos 0
	set leader {}
	# Vince's improvement (but doesn't allow embedded braces):
	# This is better because if you have a brace in the mark-menu you can't
	# get rid of it!
 	set exp {\\((sub)*section|chapter|input|include|usepackage)(\[.*\]|\*)?{[^{}]*}}
#	set exp {\\((sub)*section|chapter|input|include|usepackage)(\[.*\]|\*)?{(.*)}}
	while { ![catch {search -s -f 1 -r 1 -m 0 -i 0 $exp $pos} res] } {
		set start [lindex $res 0]
		set end [lindex $res 1]
		set cmd [getText $start $end]
		if { [regexp {\{(.*)\}} $cmd dummy arg] } {
			set item ""
			set cmdName [extractCommandName $cmd]
			if { [regexp {((sub)*)section} $cmdName dummy prefix] } {
				# indent \(sub)*section or \(sub)*section*:
				set spaces [expr 2 * [string length $prefix] / 3]
				append item $leader [format "%$spaces\s" ""]
			} elseif { [regexp {chapter} $cmdName] } {
				# if \chapter or \chapter*, indent next \(sub)*section:
				set leader {  }
			} else {
				# mark \input, \include, and \usepackage (VD):
				append item ""
			}
			# remove all superfluous whitespace (WTP):
			regsub -all "\[\ \r\t\]\+" $arg { } arg
			# limit the width of the menu item (WTP):
			if { [string length $arg] > 30 } { set arg "[string range $arg 0 26]" }
			# build the menu item:
			append item $arg
			# create the mark:
			setNamedMark $item [lineStart [expr $start - 1]] $start $start
		}
		set pos [expr $end + 1]
	}
}

proc TeX::Sty_MarkFile {} {
	set pos 0
	set leader {}
 	set exp {\\((re)?newcommand|def|DeclareRobustCommand)\\([^[\#\\\{\r\n%]*)}
	while { ![catch {search -s -f 1 -r 1 -m 0 -i 0 $exp $pos} res] } {
		set start [lindex $res 0]
		set end [lindex $res 1]
		set cmd [getText $start $end]
		if { [regexp {\\((re)?newcommand|def|DeclareRobustCommand)\\([^[\#\\\{\r\n%]*)} $cmd d d d item] } {
			# limit the width of the menu item (WTP):
			if { [string length $item] > 30 } { set item "[string range $item 0 26]" }
			# save the mark:
			lappend items [list $item $start]
		}
		set pos [expr $end + 1]
	}
	foreach i [lsort -ignore $items] {
		set start [lindex $i 1]
		setNamedMark [lindex $i 0] [lineStart [expr $start - 1]] $start $start
	}
}

proc TeX::parseFuncs {} {
	global funcExpr parseExpr

	set m {}
	set pos 0
	while {[set res [search -s -f 1 -r 1 -i 0 -n $funcExpr $pos]] != ""} {
		set text [getText [car $res] [expr [nextLineStart [cadr $res]] - 1]]
		if {[regexp $parseExpr $text dummy word]} {
			set num [regsub -all sub $text sub dummy]
			lappend m "[format %${num}s {}]$word" [car $res]
		}
		set pos [cadr $res]
	}
	return $m
}

#--------------------------------------------------------------------------
# Command-double-clicking:
#--------------------------------------------------------------------------
# In TeX mode, use cmd-double-clicks to follow references and citations,
# or open input files.
#
# (originally written by Tom Pollard and Andreas Amann)
#
# Remaining bugs:
#   - search is successful even if the pattern is commented out
#     (which is a bug or a feature, depending how you look at it)
	
# Extended to work with filesets and with better bib code
proc TeX::DblClick {from to} {
	global TeXmodeVars

	# Extend selection to largest string delimited by commas or curly-braces
	set text [string trim [eval getText [TeXExtendArg $from $to]]]

	# Set $cmd to TeX command for which the selection is an argument, but
	# only if user clicked on a valid command argument.
	set cmd {}
	if {[set mtch [findCommandWithParts $from 0]] != ""} {
		set beg [lindex $mtch 0]
		set arg [lindex $mtch 2]
		set end [lindex $mtch 3]
		# Make sure the user clicked within a TeX argument and not
		# on the command name itself
		if {$from > $arg && $to < $end} {
			set cmd [extractCommandName [getText $beg $arg]]
		}
	}
	
	switch $cmd {
	  "@input" {openTeXFile $text ".tex"}
	  "input" {openTeXFile $text ".tex"}
	  "include" {openTeXFile "${text}.tex"}
	  "documentclass" -
	  "LoadClass" {openTeXFile "${text}.cls"}
	  "usepackage" -
	  "RequirePackage" {openTeXFile "${text}.sty"}
	  "bibliography" {bib_OpenFile "${text}.bib"}
	  "bibliographystyle" {bib_OpenFile "${text}.bst"}
	  default {
			# \cite, \nocite, etc.
			if {[lsearch -exact $TeXmodeVars(citeCommands) $cmd] >= 0} {
				TeX_dblClickCitation $text
				return
				
			# \ref, \pageref, etc.
			} elseif {[lsearch -exact $TeXmodeVars(refCommands) $cmd] >= 0} {
				set labelPat {\\label\{}
				# Check for \label in current window/fileset
				if [TeX_selectPatternInFileOrSet "$labelPat$text\}" "tex"] {
					return
				}
				# Only gets here if the search failed
				beep
				popPosition
				message {No matching \label found}
			# box-making macro ($boxMacroNames)
			} elseif { [lsearch -exact $TeXmodeVars(boxMacroNames) $cmd] > -1 } {
				openTeXFile $text 
				
			# Other
			} else {
				select $from $to
				message {Command-double-click on the required argument of an underlined LaTeX command}
			}
		}
	}
}

## 
 # -------------------------------------------------------------------------
 # 
 # "TeX_selectPatternInFileOrSet" --
 # 
 #  Useful procedure which will search for a given pattern in all files in
 #  the current fileset (or the fileset of the given window) if such a 
 #  fileset exists, and otherwise it will just look at the current tex
 #  window.  The 'types' argument may be '*', 'bib' or 'tex' depending
 #  upon which types of file you wish to search in the fileset.
 #  
 #  Returns 1 if successful, else 0.
 # -------------------------------------------------------------------------
 ##
proc TeX_selectPatternInFileOrSet {pat {types "*"} {win ""}} {
	if {$win == ""} { set win [win::Current] }
	if {[set fset [isWindowInFileset $win "tex"]] == ""} {
		if [selectPatternInFile $win "$pat"] {
			return 1
		}
	} else {
		foreach filename [texListFilesInFileSet $fset $types] {
			if [selectPatternInFile $filename "$pat"] {
				return 1
			}			
		}
	} 
	return 0
}

proc TeX_dblClickCitation {text} {
	global TeXSearchPath
	# Quote regexp-active characters in case we use $text in a regexp search
	set qtext [quote::Regfind $text]
	set bibPat "\\bibliography\{"
	set theBibPat "\\begin\{thebibliography"
	set bibitemPat {\bibitem(\[[^]]*\])?\{}
	set bibTopPat {@([a-zA-Z]+)[\{\(][ 	]*}
	
	# Check first in the current window, either for a bib entry or a \bibitem
	if [TeX_findBibItem $qtext [list [win::Current]]] {
		return
	}
	# If we're in a fileset...
	if {[set theFileset [isWindowInFileset "" "tex"]] != ""} {
		# search all the files in that fileset for bib entries or \bibitems
		if [TeX_findBibItem $qtext [texListFilesInFileSet $theFileset "bib"]] { 
			return
		}
		if [TeX_findBibItem $qtext [texListFilesInFileSet $theFileset "tex"]] { 
			return
		}
		beep
		message "can't find \"$text\" in the .bib file(s) or in \\thebibliography"
	} else {
		# if no fileset, and we didn't find it above, look for all the
		# included bibliographies, and search them:
		if {![catch {search -f 0 -r 0 -s -m 0 $bibPat [maxPos]} mtch]} {
			# Get ALL the bib file names:
			set beg [lindex $mtch 1]
			set end [matchIt "\{" $beg]
			set bibnames "[split [getText $beg $end] ,]"
			# Check every file:
			foreach fname $bibnames {
				if [selectPatternInFile [absolutePath "${fname}.bib"] $bibTopPat$qtext] {
					return
				}
			}
			# Build the TeX search path:
			if { [llength $TeXSearchPath] == 0 } {
				message "building TeX search path"
				set TeXSearchPath [buildTeXSearchPath]
				message ""
			}
			# Check every file in the TeX search path:
			foreach fname $bibnames {
				foreach folder $TeXSearchPath {
					if [selectPatternInFile "$folder${fname}.bib" $bibTopPat$qtext] {
						return
					}
				}
			}
			beep
			message "can't find \"$text\" in the .bib file(s)"	
		} else {
			beep
			message "can't find a \\bibliography"
		}
	}
	# Search through all the bibliographies using the bibIndex
	# (this requires Vince's Additions).  If we find the entry,
	# prompt to insert a new database in the list of bibliographies.
	set c [TeX_currentBaseFile [win::Current]]
	if {![catch {bib_GotoEntryFromIndex $text}]} {
		if {[askyesno "This entry is in a bibliography file which isn't referenced.  Shall I add it to this document's list of bibliographies?"] == "yes"} {
			bib_insertNewBibliography $c [file::baseName [win::Current]]
		}
		return
	}
	# There's no matching entry.  Give a list of possibilities.
	bib_NoEntryExists $text [TeX_currentBaseFile]
}

proc TeX_findBibItem {qtext fnames} {
	set theBibPat "\\begin\{thebibliography"
	set bibitemPat {\bibitem(\[[^\]*\])?\{}
 	set bibTopPat {@([a-zA-Z]+)[\{\(][ 	]*}
	foreach filename $fnames {
		if ![file exists $filename] { continue }
		switch [file extension $filename] {
			".tex" -
			".ltx" {
				# Check first for a \thebibliography environment...
				set searchResult [searchInFile $filename $theBibPat 1]
				if {[set pos [lindex $searchResult 0]] >= 0} {
					# ...then for the \bibitem, which must be after
					# \thebibliography declaration to make any sense
					if [selectPatternInFile $filename "$bibitemPat$qtext\}" $pos] {
						return 1
					}
				}
			}
			".bib" {
				if [selectPatternInFile $filename $bibTopPat$qtext] {
					return 1
				}
			}
		}
	}
	return 0
}

proc TeX::RequirePackage {pkg} {
	# search the document/fileset to ensure we have a 'usepackage'
	# or 'RequirePackage' command for this package.
	
	set filename [TeX_currentBaseFile]
	set pat {\\(usepackage|RequirePackage)}
	append pat "\{$pkg\}"
	if {[searchInFile $filename $pat 0] == ""} {
		# search through sub-packages
		set pat2 {^.*\\(usepackage|RequirePackage)\{([^\}]+)\}(.*)$}
		set subpkgs [file::findAllInstances $filename $pat2 1]
		foreach f $subpkgs {
			if ![catch {set sty [findTeXFile $f ".sty"]}] {
				if {[searchInFile $sty $pat 0] != ""} { return }
			}
		}
		global TeXmodeVars
		if $TeXmodeVars(warnIfPackageNotIncluded) {
			beep
			if {[askyesno "The '$pkg' package seems not to be included in this LaTeX document.  Shall I insert a 'usepackage' line?"] == "yes"} {
				set w [win::Current]
				pushPosition
				TeX::InsertInPreamble "\\usepackage\{$pkg\}\r"
				bringToFront $w
				popPosition
			}
		}
	}
}

proc TeX::InsertInPreamble {text} {
	if [selectPatternInFile [TeX_currentBaseFile] {\\documentclass(\[[^]]*\])?\{[^\}]*\}} ] {
		goto [nextLineStart [selEnd]]
		insertText $text
		return 1
	} else {
		alertnote "Can't find the document preamble '\\documentclass...'"
		return 0
	}
}

proc TeX::Format {menu item} {TeX::SetFormat [lindex $item 0]}
proc TeX::SetFormat {text} {
	openFileQuietly [TeX_currentBaseFile]
	if [regexp {^%&[^ \t]*$} [getText 0 [set p [nextLineStart 0]]]] {
		replaceText 0 $p "%&$text\r"
	} else {
		goto 0
		insertText "%&$text\r"
	}
	select 0 [nextLineStart 0]
}

## 
 # -------------------------------------------------------------------------
 #	 
 # "TeXOptionTitlebar"	--
 #	
 #	Invoked	by option-clicking on the title-bar.  List either all files	in
 #	the	current	fileset	(if	we're part of one),	or just	all	input/include
 #	files.	Select one to open it in a window.
 # -------------------------------------------------------------------------
 ##
proc TeX::OptionTitlebar {} {
	set win [win::Current]
	if {[set fset [isWindowInFileset $win "tex"]] == ""} {
		# find includes directly
		set cid [scancontext create]
		set lines {}
		# copes with various commands with '[]' optional
		global TeX::subFileCommmands
		scanmatch $cid "^\[ \t\]*\\\\([join ${TeX::subFileCommmands} |])(\\\[\[^\]\]*\\\])?\{(.*)\}"  {
			#eval lappend lines [split $matchInfo(submatch2) ,]
			lappend lines [list $matchInfo(submatch0) [split $matchInfo(submatch2) ,]]
		}
		set fid [open $win "r"]
		scanfile $cid $fid
		close $fid
		scancontext delete $cid
		if {$lines != ""} {
			foreach l $lines {
				set type [lindex $l 0]
				foreach f [lindex $l 1] {
					switch $type {
						"bibliography" {
							lappend files "${f}.bib"
						}
						"bibliographystyle" {
							lappend files "${f}.bst"
						}
						"usepackage" {
							lappend files "${f}.sty"
						}
						"RequirePackage" {
							lappend files "${f}.sty"
						}
						default {
							lappend files "${f}"
						}
					}
				}
			}
			return $files
		} else {
			# just use files in this directory instead, so signal error
			error ""
		}
		
	} else {
		set lines {}
		global fileSetsExtra
		foreach f $fileSetsExtra($fset) { 
			if {[lindex $f 0] == "tex"} {
				lappend lines "[string range {      } 1 [expr 2*[lindex $f 1]]][lindex $f 2]" 
			}
		}
		return $lines
	}
}
# used by the above proc
set TeX::subFileCommmands [list \
	include usepackage RequirePackage input bibliography bibliographystyle \
]


## 
 # -------------------------------------------------------------------------
 #	 
 # "TeXOptionTitlebarSelect" --
 #	
 #	Procedure called when you select an	item from the option pop-up	menu.
 #	We just	open the file if we	can.  If there isn't a fileset,	this
 #	procedure could	fail.
 # -------------------------------------------------------------------------
 ##
proc TeX::OptionTitlebarSelect {fname} {
	set fname [string trim $fname]
	if {[set fset [isWindowInFileset "" "tex"]] == ""} {
		switch [file extension $fname] {
			".bib" -
			".bst" {
				bib_OpenFile $fname
			}
			default {
				openTeXFile $fname
			}
		}
	} else {
		openFileQuietly [texfs_awkwardGetFile $fset $fname]
	}
}

proc TeX_currentBaseFile {{currentWin ""}} {
	if {$currentWin == ""} {
		set currentWin [win::Current]
	}
	if {[set currentProj [isWindowInFileset $currentWin "tex"]] !=	""} {
		return [texFilesetBaseName $currentProj]
	} else {
		return $currentWin
	}
}
## 
 # -------------------------------------------------------------------------
 #	 
 # "TeXEnsureSearchPathSet"	--
 #	
 #	Make sure TeX mode has built our search	path, so we	can	find 
 #	bibliography files.	 Perhaps we	should have	our	own	variable
 #	for	these?
 # -------------------------------------------------------------------------
 ##
proc TeXEnsureSearchPathSet {} {
	global TeXSearchPath
	if { [llength $TeXSearchPath] == 0 } {
		message "building TeX search path"
		set TeXSearchPath [buildTeXSearchPath]
		message ""
	}
}


## 
 # -------------------------------------------------------------------------
 # 
 # "findTeXFile" --
 # 
 #  Find a TeX file with given default extension if it exists anywhere on
 #  the TeX search path.
 # -------------------------------------------------------------------------
 ##
proc findTeXFile {file {ext ""}} {
	global TeXSearchPath
	# Determine absolute file specification
	# Ignore $ext if $file already has an extension
	if {[string length [file extension $file]] == 0} {
		set file $file$ext
	}
	set filename [absolutePath $file]
	if [file exists $filename] { 
		return $filename
	}
	if { [llength $TeXSearchPath] == 0 } {
		message "building TeX search path"
		set TeXSearchPath [buildTeXSearchPath]
		message ""
	}
	foreach folder $TeXSearchPath {
		set filename "$folder$file"
		if [file exists $filename] {
			return $filename
		}
	}
	error "File not found."
}
proc openTeXFile {file {ext ""}} {
	if [catch {set f [findTeXFile $file $ext]}] {
		beep
		error "can't find TeX input file \"$file\""
	} else {
		openFileQuietly $f
		message $f
	}
}

# Return a list of folders in which to search for TeX input files, 
# including the TeXInputs folder (if it exists) and any folders of 
# the form "*input*" in the TeX application directory.  The current
# folder is not included in the list.  Default search depth is two
# levels deep.
#
# (Note: The TeXInputs folder is assigned from the AppPaths submenu.)
#
proc buildTeXSearchPath {{depth 2}} {
	global TeXmodeVars texSig
	
	set folders {}
	# The local 'TeXInputsFolder' folder:
	if {[info exists TeXmodeVars(TeXInputsFolder)] && \
	  [string length $TeXmodeVars(TeXInputsFolder)] > 0} { 
		set folders [concat $folders [list $TeXmodeVars(TeXInputsFolder)]]
		# Search subfolders $depth levels deep:
		set folders [concat $folders [listSubfolders $TeXmodeVars(TeXInputsFolder) $depth]]
	}

	# Any "*inputs*" folders in the TeX application folder:
	if {[info exists texSig] && [string length $texSig] > 0} { 
		set texPath [nameFromAppl $texSig]
		set TeXDir [file dirname $texPath]
		# Why add the folder containing the TeX app to the list?
		# set folders [concat $folders [list $TeXDir]]
		# Bug:  'glob' is case sensitive!
		foreach folder [glob "$TeXDir:*\[Ii\]nputs*"] {
			set folders [concat $folders [list $folder]]
			# Search subfolders $depth levels deep:
			set folders [concat $folders [listSubfolders $folder $depth]]
		}
	}

	# Make sure each folder ends with a colon:
	set result {}
	foreach folder $folders {
		set folder "[string trimright $folder {:}]:"
		set result [concat $result [list $folder]]
	}
	return $result
}


# Extend the argument around the position $from.
# (Args are delimited by commas or curly-braces.)
proc TeXExtendArg {from {to 0}} {
	if {$to == 0} { set to $from }
	set result [list $from $to]
	if {![catch {search -f 0 -r 1 -s -m 0 "\[,\{\]" $from} mtch0]} {
		if {![catch {search -f 1 -r 1 -s -m 0 "\[,\}\]" $to} mtch1]} {
			set from [lindex $mtch0 1]
			set to [lindex $mtch1 0]
			## Embedded braces in the arg probably mean that the user
			## clicked outside a valid command argument
			if {[regexp "\[\{\}\]" [getText $from $to]] == 0} {
				set result [list $from $to]
			}
		}
	}
	return $result
}

# Find a LaTeX command with arguments in either direction.
# (see findCommandWithArgs in latexMacros.tcl)
# This version returns the positions at which the command options 
# and arguments start, as well.
proc findCommandWithParts {pos direction} {
	set searchString {\\([^@a-zA-Z\t\r]|@?[a-zA-Z]+\*?)(\[.*\])*({[^{}]*})?}
	if {![catch {search -s -f $direction -r 1 $searchString $pos} mtch]} {
		set beg [lindex $mtch 0]
		set end [lindex $mtch 1]
		regexp -indices $searchString [getText $beg $end] all cmd opt arg
		set opt [expr $beg + [lindex $opt 0]]
		set arg [expr $beg + [lindex $arg 0]]
		return [list $beg $opt $arg $end]
	} else {
		return ""
	}
}

#--------------------------------------------------------------------------
# Insertion routines:
#--------------------------------------------------------------------------

# Returns a modified text string if the string $text is non-null, 
# and the null string otherwise.  The argument 'operation' is a 
# string directing 'doPrefixText' to either "insert" or "remove" 
# $prefixString to/from each line of $text.  (This routine is
# adapted from strings.tcl.)
# proc doPrefixText {operation prefixString text} {
# 	if {$text == ""} {return ""}
# 	set pref [quote::Regfind $prefixString]
# 	if {$operation == "insert"} {
# 		set trailChar ""
# 		set textLen [string length $text]
# 		if {[string index $text [expr $textLen-1]] == "\r"} {
# 			set text [string range $text 0 [expr $textLen-2]]
# 			set trailChar "\r"
# 		}
# 		set str \r$prefixString
# 		regsub -all \r $text $str text
# 		return $prefixString$text$trailChar
# 	} elseif {$operation == "remove"} {
# 		regsub -all \r$pref $text \r text
# 		regsub ^$pref $text "" text
# 		return $text
# 	}
# }

# Shift each line of $text to the right by inserting a string of
# $whitespace characters at the beginning of each line, returning
# the resulting string.
proc shiftTextRight {text whitespace} {
	return [doPrefixText "insert" $whitespace $text]
}

# Return a string of whitespace characters from the beginning 
# of the line containing $pos up to the first non-whitespace 
# character.
proc getIndentation {pos} {
	set text [getText [lineStart $pos] [nextLineStart $pos]]
	regexp {^[ \t]*} $text theIndentation
	return $theIndentation
}

# Return an "indented carriage return" if any character preceding 
# the insertion point (on the same line) is a non-whitespace 
# character.  Otherwise, return the null string.
proc openingCarriageReturn {} {
	set pos [getPos]
	set end $pos
	set start [lineStart $pos]
	set text [getText $start $end]
	if {[isWhitespace $text]} {
		if {$start == $end} {return [getIndentation $pos]} {return ""}
	} else {
		return "\r[getIndentation $pos]"
	}
}
# Return an "indented carriage return" if any character following 
# the insertion point (on the same line) is a non-whitespace 
# character.  Otherwise, return the null string.
proc closingCarriageReturn {} {
	set pos [selEnd]
	if {[isSelection] && ($pos == [lineStart $pos])} {
		return "\r"
	} else {
		set start $pos
		set end [nextLineStart $start]
		set text [getText $start $end]
		if {[isWhitespace $text]} {
			return ""
		} else {
			return "\r[getIndentation $pos]"
		}
	}
}

# Insert an object at the insertion point. If there is a selection and the 
# global variable 'deleteObjNoisily' is false, quietly delete the selection 
# (like 'paste'). Otherwise, prompt the user for the appropriate action. 
# Returns true if the object is ultimately inserted, and false if the 
# user cancels the operation. 
proc insertObject {objectName} {
	global TeXmodeVars
	if {[isSelection]} {
		if {$TeXmodeVars(deleteObjNoisily)} {
			switch [askyesno -c "Delete selection?"] {
				"yes" {deleteText [getPos] [selEnd]}
				"no" {backwardChar}
				"cancel" {return 0}
			}
		} else {
			deleteText [getPos] [selEnd]
		}
	}
	elec::Insertion $objectName
	return 1
}

# Insert an object at the insertion point. If there is a selection, wrap 
# it inside the parameters $left and $right. Returns true if there is a 
# selection (in which case it will wrap), and false otherwise. 
proc wrapObject {left right {complete 0}} {
	set selected [isSelection]
	if {$selected} {
		elec::ReplaceText [getPos] [selEnd] $left [getSelect] $right
	} else {
		elec::Insertion $left "" $right
		if $complete {
			bind::Completion
		}
	}
	return $selected
}

# Builds and returns a LaTeX environment, that is, a \begin...\end 
# pair, given the name of the environment, an argument string, 
# and the environment body.  The body should be passed to this 
# procedure fully formatted, including indentation.
# proc buildEnvironment {envName envArg envBody trailingComment} {
# 	set indent [getIndentation [getPos]]
# 	set envStr [openingCarriageReturn]
# 	append envStr "\\begin{" $envName "}"
# 	append envStr $envArg "\r"
# 	append envStr $envBody
# 	append envStr $indent "\\end{" $envName "}$trailingComment"
# 	append envStr [closingCarriageReturn]
# 	return $envStr
# }

# Builds and returns a LaTeX environment, that is, a \begin...\end 
# pair, given the name of the environment, an argument string, 
# and the environment body.  The body should be passed to this 
# procedure fully formatted, including indentation.
proc buildEnvironment {envName envArg envBody trailingComment} {
	append begStruct "\\begin{" $envName "}" $envArg
	append endStruct "\\end{" $envName "}$trailingComment"
	return [buildStructure $begStruct $envBody $endStruct]
}
# Builds and returns a fully-formed structure, a string of the form
#
#   <begStruct>
#     <bodyStruct>
#   <endStruct>
#
# For example,
#
#   buildStructure "if {} {" "\t\r" "}"
#
# returns a Tcl if-template.
#
proc buildStructure {begStruct bodyStruct endStruct} {
	set indent [getIndentation [getPos]]
	set structure [openingCarriageReturn]
	append structure $begStruct "\r"
	append structure $bodyStruct
	append structure $indent $endStruct
	append structure [closingCarriageReturn]
	return $structure
}

# Inserts a LaTeX environment with the specified name, argument, 
# and body at the insertion point.  Deletes the current selection 
# quietly if the global variable 'deleteEnvNoisily' is false; 
# otherwise the user is prompted for directions.  Returns true if the 
# environment is ultimately inserted, and false if the user cancels 
# the operation.
proc insertEnvironment {envName envArg envBody} {
	global TeXmodeVars
	if { [isSelection] } {
		if { $TeXmodeVars(deleteEnvNoisily) } {
			switch [askyesno -c "Delete selection?"] {
				"yes" {}
				"no" {backwardChar}
				"cancel" {return 0}
			}
		}
	}
	append begStruct "\\begin{" $envName "}" $envArg
	append endStruct "\\end{" $envName "}"
	insertStructure $begStruct $envBody $endStruct
	return 1
}
# Inserts a structure at the insertion point.  Positions the cursor 
# at the beginning of the structure, leaving any subsequent action 
# to the calling procedure.  Deletes the current selection quietly.
proc insertStructure {begStruct bodyStruct endStruct} {
	set start [getPos]
	set end [selEnd]
	set body [shiftTextRight $bodyStruct [getIndentation $start]]
	elec::ReplaceText $start $end [buildStructure $begStruct $body $endStruct]
}


# Inserts an environment with the given name, argument, and body at 
# the insertion point.  If there is currently a selection, cut and 
# paste it into the body of the new environment, maintaining proper 
# indentation; otherwise, insert a tab stop into the body of the
# environment.  Returns true if there is a selection, and false 
# otherwise.
proc wrapEnvironment {envName envArg envBody} {
	append begStruct "\\begin{" $envName "}" $envArg
	append endStruct "\\end{" $envName "}"
	return [wrapStructure $begStruct $envBody $endStruct]
}
# Inserts a structure at the insertion point.  Positions the cursor 
# at the beginning of the structure, leaving any subsequent action 
# to the calling procedure.  If there is currently a selection, cut 
# and paste it into the body of the new environment, maintaining proper 
# indentation; otherwise, insert a tab stop into the body of the
# environment.  Returns true if there is a selection, and false 
# otherwise.
proc wrapStructure {begStruct bodyStruct endStruct} {
	set start [getPos]
	set end [selEnd]
	set indent [getIndentation $start]
	if [isSelection] {
		set text [getSelect]
		set textLen [string length $text]
		if { [string index $text [expr $textLen-1]] != "\r" } {
			append text "\r"
		}
		if {$start == [lineStart $start]} {
			set body [shiftTextRight $text \t]
		} else {
			set body "$indent[shiftTextRight $text \t]"
		}
		append body [shiftTextRight $bodyStruct $indent]
		set returnFlag 1
	} else {
		append body "$indent\t\r" [shiftTextRight $bodyStruct $indent]
		set returnFlag 0
	}
	elec::ReplaceText $start $end [buildStructure $begStruct $body $endStruct]
	return $returnFlag
}

# A generic call to 'wrapEnvironment' used throughout latex.tcl:
proc doWrapEnvironment {envName} {
	if [wrapEnvironment $envName "" ""] {
		set msgText "selection wrapped"
	} else {
		set msgText "enter body of $envName environment"
	}
	message $msgText
}
# A generic call to 'wrapStructure':
proc doWrapStructure {begStruct bodyStruct endStruct} {
	if [wrapStructure $begStruct $bodyStruct $endStruct] {
		set msgText "selection wrapped"
	} else {
		set msgText "enter body of structure"
	}
	message $msgText
}

# Inserts a structured document template at the insertion point.  
# Three arguments are required:  the class name of the document, a 
# preamble string, and a string containing the body of the document.  
# If the preamble is null, a generic \usepackage statement is 
# inserted; otherwise, the preamble is inserted as is.  This routine 
# does absolutely no error-checking (this is totally left up to the 
# calling procedure) and returns nothing.
proc insertDocument {className preamble docBody} {
	set docStr "\\documentclass\[\]{$className}\r"
	if {$preamble == ""} {
		append docStr "\\usepackage\[\]{}\r\r\r\r"
	} else {
		append docStr $preamble
	}
	append docStr [buildEnvironment "document" "" $docBody "\r"]
	set start [getPos]
	set end [selEnd]
	elec::ReplaceText $start $end $docStr
	return
}

# Inserts a document template at the insertion point given the 
# class name of the document to be inserted.  If ALL of the current
# document is selected, then the routine wraps the text inside a
# generic document template.  If the file is empty, a bullet is 
# inserted in place of the document body.  If neither of these 
# conditions is true, no action is taken.  Returns true if 
# wrapping occurs, and false otherwise.
proc wrapDocument {className} {
	global TeXmodeVars
	if { [isEmptyFile] } {
		append body "\r\r\r"
# 		set returnFlag 0
	} else {
		if {[isSelectionAll]} {
			set text [getSelect]
			append body "\r$text\r"
# 			set returnFlag 1
		} else {
			if { $TeXmodeVars(searchNoisily) } {beep}
			alertnote "nonempty file:  delete text or \'Select All\'\
				from the Edit menu"
			return 0
		}
	}
	set docStr "\\documentclass\[\]{$className}\r"
	append docStr "\\usepackage\[\]{}\r\r\r\r"
	append docStr [buildEnvironment "document" "" $body "\r"]
	set start [getPos]
	set end [selEnd]
	elec::ReplaceText $start $end $docStr
# 	return $returnFlag
	return 1
}

#--------------------------------------------------------------------------
# Booleans to determine the location of the insertion point
#--------------------------------------------------------------------------

# Return true if the insertion point is before the preamble, and 
# false otherwise.   Define "before the preamble" to be all text to 
# the left of "\" in "\documentclass".
proc isBeforePreamble {} {
	set searchString "\\documentclass"
	set searchResult [search -s -f 1 -r 0 -n $searchString [getPos]]
	if {[llength $searchResult]} {
		return 1
	} else {
		return 0
	}
}

# Return true if the insertion point is in the preamble, and false
# otherwise.  Define "preamble" to be all text to the left of "\" 
# in "\begin{document}", but not before the "\" in "\documentclass".
proc isInPreamble {} {
	set searchString "\\begin\{document\}"
	set searchResult [search -s -f 1 -r 0 -n $searchString [getPos]]
	if {[llength $searchResult]} {
		return 1
	} else {
		return 0
	}
}

# Return true if the insertion point is in the document environment,
# and false otherwise.  Define "document" to be all text between 
# "\begin{document}" and "\end{document}", exclusive.
proc isInDocument {} {
	set pos [getPos]
	set searchString "\\begin\{document\}"
	# adjust for the length of the search string:
	set len [string length $searchString]
	if {$pos < $len} {return 0}
	set searchResult [search -s -f 0 -r 0 -n $searchString [expr $pos - $len]]
	if {[llength $searchResult]} {
		set searchString "\\end\{document\}"
		set searchResult [search -s -f 1 -r 0 -n $searchString $pos]
		if {[llength $searchResult]} {
			return 1
		} else {
			return 0
		}
	} else {
		return 0
	}
}

# Return true if the insertion point is after the document
# environment, and false otherwise.  Define "after the document 
# environment" to be all text to the right of "}" in "\end{document}".
proc isAfterDocument {} {
	set pos [getPos]
	set searchString "\\end\{document\}"
	set searchResult [search -s -f 0 -r 0 -n $searchString $pos]
	if {[llength $searchResult]} {
		set pos1 [lindex $searchResult 0]
		set pos2 [lindex $searchResult 1]
		if {$pos1 <= $pos && $pos < $pos2} {
			return 0
		} else {
			return 1
		}
	} else {
		return 0
	}
}

proc isInMathMode {} {
	set pos [getPos]
	# Check to see if in LaTeX math mode:
	set searchString {\\[()]}
	set searchResult1 [search -s -f 0 -r 1 -n $searchString [expr $pos - 1]]
	if {[llength $searchResult1]} {
		set delim1 [eval getText $searchResult1]
	} else {
		set delim1 "none"
	}
	set searchResult2 [search -s -f 1 -r 1 -n $searchString $pos]
	if {[llength $searchResult2]} {
		set delim2 [eval getText $searchResult2]
	} else {
		set delim2 "none"
	}
	set flag1 [expr [string match {none} $delim1] && [string match {\\)} $delim2]]
	set flag2 [expr [string match {\\(} $delim1] && [string match {none} $delim2]]
	set flag3 [expr [string match {\\(} $delim1] && [string match {\\)} $delim2]]
	set flag4 [expr [string match {\\(} $delim1] && [string match {\\(} $delim2]]
	set flag5 [expr [string match {\\)} $delim1] && [string match {\\)} $delim2]]
	if {$flag3} {
		return 1
	} elseif {$flag1 || $flag2 || $flag4 || $flag5} {
		set messageString "unbalanced math mode delimiters"
		beep
		alertnote $messageString
		error "isInMathMode:  $messageString"
	}
	# Check to see if in LaTeX displaymath mode:
	set searchString {[^\\]\\\[|\\\]}
	set searchResult1 [search -s -f 0 -r 1 -n $searchString [expr $pos - 1]]
	if {[llength $searchResult1]} {
		set begPos [lindex $searchResult1 0]
		set endPos [lindex $searchResult1 1]
		if {[lookAt [expr $endPos - 1]] == "\["} {
			set delim1 [getText [expr $begPos + 1] $endPos]
		} else {
			set delim1 [eval getText $searchResult1]
		}
	} else {
		set delim1 "none"
	}
	set searchResult2 [search -s -f 1 -r 1 -n $searchString $pos]
	if {[llength $searchResult2]} {
		set begPos [lindex $searchResult2 0]
		set endPos [lindex $searchResult2 1]
		if {[lookAt [expr $endPos - 1]] == "\["} {
			set delim2 [getText [expr $begPos + 1] $endPos]
		} else {
			set delim2 [eval getText $searchResult2]
		}
	} else {
		set delim2 "none"
	}
	set flag1 [expr [string match {none} $delim1] && [string match {\\\]} $delim2]]
	set flag2 [expr [string match {\\\[} $delim1] && [string match {none} $delim2]]
	set flag3 [expr [string match {\\\[} $delim1] && [string match {\\\]} $delim2]]
	set flag4 [expr [string match {\\\[} $delim1] && [string match {\\\[} $delim2]]
	set flag5 [expr [string match {\\\]} $delim1] && [string match {\\\]} $delim2]]
	if {$flag3} {
		return 1
	} elseif {$flag1 || $flag2 || $flag4 || $flag5} {
		set messageString "unbalanced math mode delimiters"
		beep
		alertnote $messageString
		error "isInMathMode:  $messageString"
	}
	# Check to see if in math environment:
	set envName [extractCommandArg [eval getText [searchEnvironment]]]
	set envList [list "math" "displaymath" "equation" "eqnarray" "eqnarray*" "array"]
	if {[lsearch -exact $envList $envName] == -1} {return 0} {return 1}
}

#--------------------------------------------------------------------------
# Routines to dissect LaTeX commands
#--------------------------------------------------------------------------

# Given a LaTeX command string, extract and return the command name.
proc extractCommandName {commandStr} {
	if {[regexp {\\([^@a-zA-Z\t\r]|@?[a-zA-Z]+\*?)} $commandStr dummy commandName]} {
		return $commandName
	} else {
		return ""
	}
}

# Given a LaTeX command string with at most one required argument and
# no embedded carriage returns, extract and return the argument.
# (Note:  this proc needs more testing.)
proc extractCommandArg {commandStr} {
	if {[regexp {\{(.*)\}} $commandStr dummy arg]} {
		return $arg
	} else {
		return ""
	}
}

#--------------------------------------------------------------------------
# An environment search routine
#--------------------------------------------------------------------------

# Search for the closest surrounding environment and return a list 
# of two positions if the environment is found; otherwise return the 
# empty list.  Assumes the LaTeX document is syntactically correct.   
# The command 'eval select [searchEnvironment]' selects the "\begin" 
# statement (with argument) of the surrounding environment.
proc searchEnvironment {} {
	watchCursor
	set pos [getPos]
	if {$pos == 0} {
		return [list]
	}
	# adjust position if insertion point is contained in "\end{...}"
	set searchPos [expr $pos - 1]
	set searchString {\\end\{[^\{\}]*\}}
	set searchResult [search -s -f 0 -r 1 -n $searchString $searchPos]
	if {[llength $searchResult]} {
		set pos1 [lindex $searchResult 0]
		set pos2 [lindex $searchResult 1]
		if {$pos1 < $pos && $pos < $pos2} {
			set searchPos [expr $pos1 - 1]
		}
	}
	# begin reverse search:
	set searchString {\\(begin|end)\{[^\{\}]*\}}
	set searchResult [search -s -f 0 -r 1 -n $searchString $searchPos]
	if {[llength $searchResult]} {
		set text [eval getText $searchResult]
	} else {
		return [list]
	}
	set depthCounter 0
	set commandName [extractCommandName $text]
	while {$commandName == "end" || $depthCounter > 0} {
		if {$commandName == "end"} {
			set depthCounter [expr $depthCounter + 1]
		} else {
			set depthCounter [expr $depthCounter - 1]
		}
		set searchPos [expr [lindex $searchResult 0] - 1]
		set searchResult [search -s -f 0 -r 1 -n $searchString $searchPos]
		if {[llength $searchResult]} {
			set text [eval getText $searchResult]
		} else {
			return [list]
		}
		set commandName [extractCommandName $text]
	}
	return $searchResult
}

#--------------------------------------------------------------------------
# Misc:
#--------------------------------------------------------------------------

proc isInteger {str1} {
	return [regexp {^(\+|-)?[0-9]+$} [string trim $str1]]
}

proc isUnsignedInteger {str1} {
	return [regexp {^[0-9]+$} [string trim $str1]]
}

proc isPositiveInteger {str1} {
	if { [isUnsignedInteger $str1] } {
		if { ![regexp {^0+$} [string trim $str1]] } {
			return 1
		}
	}
	return 0
}


# Returns true if the entire window is selected, and false otherwise.
proc isSelectionAll {} {
	return [expr {([getPos] == 0) && ([selEnd] == [maxPos])}]
}

# Checks to see if the current window is empty, except for whitespace.
proc isEmptyFile {} {
	return [isWhitespace [getText 0 [maxPos]]]
}

# If there is a selection, make sure it's uppercase.  Otherwise, 
# check to see if the character after the insertion point is uppercase.
proc isUppercase {} {
	if {[isSelection]} {
		set text [getSelect]
	} else {
		set text [lookAt [getPos]]
	}
	return [expr {[string toupper $text] == $text}]
}

# If there is a selection, make sure it's alphabetic.  Otherwise, 
# check to see if the character after the insertion point is alphabetic.
proc isAlphabetic {} {
	if {[isSelection]} {
		set text [getSelect]
	} else {
		set text [lookAt [getPos]]
	}
	return [regexp {^[a-zA-Z]+$} $text]
}

# Takes any string and tests whether or not that string contains all 
# whitespace characters.  Carriage returns are considered whitespace, 
# as are spaces and tabs.  Also returns true for the null string.
proc isWhitespace {anyString} {
	set len [string length $anyString]
	for {set i 0} {$i < $len} {incr i} {
		set c [string index $anyString $i]
		if {($c != "\ ") && ($c != "\t") && ($c != "\r")} {return 0}
	}
	return 1
}

# Select the line containing the insertion point.
proc lineSelect {} {
	goto [lineStart [getPos]]
	nextLineSelect
}

# Delete the proc with the given proc name:
proc deleteProc {procName} {catch {rename $procName ""}}

proc checkMathMode {procName mathModeFlag} {
}

# Check to see if the LaTeX symbol package has been loaded; if not, ask 
# the user for directions.  Returns false if the package is not loaded
# AND the user cancels the operation; otherwise, returns true.
proc isSymbolPackageLoaded {} {
	global searchNoisily
	set begPos [getPos]
	set endPos [selEnd]
	set searchString {\\usepackage\{.*latexsym.*\}}
	set searchResult [search -s -n -f 0 -m 0 -i 1 -r 1 $searchString $begPos]
	if {[llength $searchResult] == 0} {
		switch [askyesno -c "Insert the LaTeX symbol package?"] {
			"yes" {
				set searchString {\\documentclass(\[.*\])?\{.*\}}
				set searchResult [search -s -n -f 0 -m 0 -i 1 -r 1 $searchString $begPos]
				if {[llength $searchResult] == 0} {
					set returnVal 0
					if {$searchNoisily} {beep}
					message "can't find \\documentclass"
				} else {
					goto [lindex $searchResult 1]
					set txt "\r\\usepackage\{latexsym\}"
					set offset [string length $txt]
					set begPos [expr $begPos + $offset]
					set endPos [expr $endPos + $offset]
					elec::Insertion $txt
					set returnVal 1
				}
			}
			"no" {set returnVal 1}
			"cancel" {set returnVal 0}
		}
	} else {
		set returnVal 1
	}
	select $begPos $endPos
	return $returnVal
}

