#############################################################################
#############################################################################
#
# latexMacros.tcl (called from latex.tcl)
#
#############################################################################
#
# Author:  Tom Scavo <trscavo@syr.edu>
#
#############################################################################
#############################################################################

proc latexMacros.tcl {} {}

#############################################################################
# 
# Basic Commands
#
#############################################################################

#--------------------------------------------------------------------------
# Utilities:
#--------------------------------------------------------------------------

# A keyboard-bound method of accessing menu commands.  Takes a list of 
# menu items (i.e., the tail of a 'menu' command), the menu name (the 
# argument of the '-n' switch) , and the name of a menu filter (the 
# argument of the '-p' switch) as input, and displays these items in a 
# list box.  If the chosen item is a menu command (as opposed to a 
# submenu), it is passed to the menu filter; otherwise, 'chooseCommand' 
# recursively calls itself until a menu command is chosen or the cancel 
# button is pressed.
#
proc chooseCommand {menuItems {menuName ""} {menuFilterProc ""} {level 1}} {
	watchCursor
	if { $menuItems == "" } { return }
	# Preprocess the list of menu items:
	foreach	item $menuItems {
		regsub -all {[<!/].} $item {} item
		regsub -all {}	$item {} item
		lappend	menOut $item
		if { [string match "menu*" $item] } {
			if { [set ind [lsearch $item {-n}]] >= 0 } {
				lappend	top "[lindex $item [incr	ind]]:"
			}
		} elseif { ![string match "(*" $item] } {
			lappend	top $item
		}
	}
	# Present the menu items to the user:
	set res	[listpick -p "Choose menu command (level $level):" $top]
	# Either execute a command or recurse on a submenu:
	if { [lsearch $menOut $res] >= 0 } {
		# Execute the command via the menu filter, if necessary:
		if { $menuFilterProc == "" } {
			$res
		} else {
			$menuFilterProc $menuName $res
		}
	} else {
		set res [string trimright $res {:}]
		foreach	item $menOut {
			if { [lsearch $item $res] >= 0 } {
				set menuItems [lindex $item end]
				# Determine the name of this submenu:
				if { [set ind [lsearch $item {-n}]] >= 0 } {
					set menuName [lindex $item [incr ind]]
				} else {
					set menuName ""
				}
				# Determine the name of the menu filter for this submenu:
				if { [set ind [lsearch $item {-p}]] >= 0 } {
					set menuFilterProc [lindex $item [incr ind]]
				} else {
					set menuFilterProc ""
				}
				return [chooseCommand $menuItems $menuName $menuFilterProc [incr level]]
			}
		}
	}
}

# A keyboard-bound method of inserting environments.
#
proc chooseEnvironment {} {
	set env [getEnvironment]
	if { [catch {eval $env}] } {
		doWrapEnvironment $env
	}
}
proc getEnvironment {} {
	if { [catch {prompt "Choose environment:" "abstract" "" \
	             "abstract" \
	             "array" \
	             "center" \
	             "description" \
	             "displaymath" \
	             "document" \
	             "enumerate" \
	             "eqnarray" \
	             "eqnarray*" \
	             "equation" \
	             "figure" \
	             "filecontents" \
	             "flushleft" \
	             "flushright" \
	             "itemize" \
	             "math" \
	             "minipage" \
	             "note" \
	             "overlay" \
	             "quotation" \
	             "quote" \
	             "slide" \
	             "table" \
	             "tabular" \
	             "thebibliography" \
	             "titlepage" \
	             "verbatim" \
	             "verse" \
            	} env] } {
		set env ""
	}
	return $env
}


#############################################################################
# 
# Paragraph Mode Macros
#
#############################################################################

#--------------------------------------------------------------------------
# Documents:
#--------------------------------------------------------------------------

proc newLaTeXDocument {} {
	catch {prompt "Choose a documentclass:" "article" "" "article" \
	        "report" "book" "letter" "slides"} documentType
	if {$documentType != "cancel"} {
		new -m TeX
		if { [catch {${documentType}Documentclass}] } {
			wrapDocument "$documentType" 
		}
		while { [options] } {}
		message "enter option (or leave blank)"
	}
}

proc letterDocumentclass {} {
	set    preamble "\r\\address\{%\r"
	append preamble "		\\\\	% insert your name here\r"
	append preamble "		\\\\	% insert your address here\r"
	append preamble "		\\\\	% insert more address here\r"
	append preamble "		  	% insert city-state-zip here\r"
	append preamble "\}\r\r"
	append preamble "\\date\{\}  % optional\r"
	append preamble "\\signature\{\}\r\r"
	set    body     "\r\\begin\{letter\}\{%\r"
	append body     "		\\\\	% insert addressee's name here\r"
	append body     "		\\\\	% insert addressee's address here\r"
	append body     "		\\\\	% insert more address here\r"
	append body     "		  	% insert addressee's city-state-zip here\r"
	append body     "\}\r\r"
	append body     "\\opening\{Dear ,\}\r\r"
	if {[isEmptyFile]} {
		append body "% BODY OF LETTER\r"
		append body "\r\r"
	} else {
		if {[isSelectionAll]} {
			set text [getSelect]
# 			deleteText 0 [maxPos]
			append body "$text\r"
		} else {
			alertnote "nonempty file:  delete text or \'Select All\'\
				from the Edit menu"
			return
		}
	}
	append body "\\closing\{Sincerely,\}\r\r"
	append body "\\encl\{\}\r"
	append body "\\cc\{\}\r\r"
	append body "\\end\{letter\}\r\r"
	insertDocument "letter" $preamble $body
	message "enter option (or leave blank)"
}
proc articleDocumentclass {} {
	if { [wrapDocument "article"] } {
		message "enter option (or leave blank)"
	}
}
proc reportDocumentclass {} {
	if { [wrapDocument "report"] } {
		message "enter option (or leave blank)"
	}
}
proc bookDocumentclass {} {
	if { [wrapDocument "book"] } {
		message "enter option (or leave blank)"
	}
}
proc slidesDocumentclass {} {
	if { [wrapDocument "slides"] } {
		message "enter option (or leave blank)"
	}
}
proc otherDocumentclass {} {
	catch {prompt "What documentclass?" "article"} documentType
	if {$documentType != "cancel"} {
		if { [wrapDocument "$documentType"] } {
			message "enter option (or leave blank)"
		}
	}
}

# If an option is inserted, return true; otherwise, return false.
proc options {} {
	set option [getOption]
	if {$option != ""} {
		insertOption $option
		return 1
	}
	return 0
}
proc getOption {} {
	catch {prompt "Choose an option:" "11pt" "" "10pt" "11pt" "12pt" "(-" \
	        "letterpaper" "legalpaper" "executivepaper" "a4paper" "a5paper" \
	        "b5paper" "(-" "landscape" "(-" "final" "draft" "(-" \
	        "oneside" "twoside" "(-" "openright" "openany" "(-" \
	        "onecolumn" "twocolumn" "(-" "notitlepage" "titlepage" \
				  "openbib" "(-" "leqno" "(-" "fleqn"} optionName
	if {$optionName != "cancel"} {
		return $optionName
	} else {
		return ""
	}
}
proc insertOption {option} {
	global TeXmodeVars
	set searchString {\\documentclass}
	set searchResult [search -s -n -f 1 -m 0 -i 1 -r 1 $searchString 0]
	if {[llength $searchResult] == 0} {
		if { $TeXmodeVars(searchNoisily) } {beep}
		message "can\'t find \\documentclass"
	} else {
		set nextCharPos [lindex $searchResult 1]
		goto $nextCharPos
		set nextChar [lookAt $nextCharPos]
		if {$nextChar == "\["} {
			forwardChar
			insertText $option
			if {[lookAt [getPos]] != "\]"} {
				insertText ","
			}
		} elseif {$nextChar == "\{"} {
			insertText "\[$option\]"
		} else {
			alertnote "unrecognizable \\documentclass statement"
		}
	}
}

proc insertPackage {package} {
	global TeXmodeVars
	# Check to see if $package is already loaded:
	if { $package != "" } {
		append searchString {^[^%]*\\usepackage\{.*} $package {.*\}}
		set searchResult [search -s -n -f 1 -m 0 -i 1 -r 1 $searchString 0]
		if {[llength $searchResult] != 0} {
			if { $TeXmodeVars(searchNoisily) } {beep}
			message "$package package already loaded"
			return
		}
	}
	# Newlines are allowed in the arguments of \documentclass:
	set searchString {\\documentclass(\[[^][]*\])?{[^{}]*}}
	# Search for \documentclass command:
	set searchResult [search -s -n -f 1 -m 0 -i 1 -r 1 $searchString 0]
	if {[llength $searchResult] == 0} {
		if { $TeXmodeVars(searchNoisily) } {beep}
		message "can't find \\documentclass"
	} else {
		pushPosition
		goto [lindex $searchResult 1]
		set txt "\r\\usepackage\{$package\}"
		insertText $txt
		backwardChar
		message "Press <Ctl .> to return to previous position"
	}
}

proc filecontents {} {
	global searchNoisily
	set searchString {\\documentclass}
	set searchResult [search -s -n -f 1 -m 0 -i 1 -r 1 $searchString 0]
	if {[llength $searchResult] == 0} {
		if {$searchNoisily} {beep}
		message "can\'t find \\documentclass"
		return
	} else {
		set prompt "File to be included:"
		if {[catch {getfile $prompt} path]} {
			return
		} else {
			replaceText 0 0 [buildFilecontents $path]
			goto 0
			message "file included"
		}
	}
}
proc filecontentsAll {} {
	global searchNoisily
	watchCursor
	message "locating all input files"
	set currentWin [win::Current]
	# Is the current window part of TeX fileset?
	set fset [isWindowInFileset $currentWin "tex"]
	if { $fset == "" } {
		set searchString {\\documentclass}
		set searchResult [search -s -n -f 1 -m 0 -i 1 -r 1 $searchString 0]
		if {[llength $searchResult] == 0} {
			if {$searchNoisily} {beep}
			message "can\'t find \\documentclass"
			return
		} else {
			set text [getText 0 [maxPos]]
		}
	} else {
		# Will not handle a base file that is open and dirty:
		set text [buildFilecontents [texFilesetBaseName $fset]]
	}
	set currentDir [file dirname $currentWin]
	set newText [texResolveAll $text $currentDir]
	if { [string length $text] == [string length $newText] } {
		beep
		message "no files to include"
	} else {
		replaceText 0 [maxPos] $newText
		goto 0
		message "all files included"
	}
}
# Takes a LaTeX document string and a path as input, and returns
# a modified document string with all filecontents environments
# prepended.
proc texResolveAll {latexDoc currentDir} {
	global TeXmodeVars
	set pairs [list \
	            {{\\documentclass} {.cls}} {{\\LoadClass} {.cls}} \
	            {{\\include} {.tex}} \
	            {{\\usepackage} {.sty}} {{\\RequirePackage} {.sty}} \
	            {{\\input} {}} \
	            {{\\bibliography} {.bib}} {{\\bibliographystyle} {.bst}} \
	          ]
	foreach macro $TeXmodeVars(boxMacroNames) {
		regsub {\*} $macro {\\*} macro
		lappend pairs [list \\\\$macro {}]
	}
	foreach pair $pairs {
		set cmd [car $pair]
		set ext [cadr $pair]
		set searchString $cmd
		append searchString {(\[[^][]*\])?{([^{}]*)}}
		set searchText $latexDoc
		while { [regexp -indices $searchString $searchText mtch dummy theArgs] } {
			set begPos [lindex $theArgs 0]
			set endPos [lindex $theArgs 1]
			set args [string range $searchText $begPos $endPos]
			foreach arg [split $args ,] {
				if { $cmd == {\\input} && ![string length [file extension $arg]] } {
					set ext {.tex}
				}
				set files [glob -nocomplain $currentDir:$arg*]
				set filename "$currentDir:$arg$ext"
				if { [lsearch -exact $files $filename] > -1 } {
					set tempDoc $latexDoc
					set latexDoc [buildFilecontents $filename]
					append latexDoc $tempDoc
				}
			}
			set searchText [string range $searchText [expr $endPos + 2] end]
		}
	}
	return $latexDoc
}
# Takes a filename as input and returns a filecontents environment 
# based on the contents of that file.  If a second argument is given,
# use that as the argument of the filecontents environment instead
# of the original filename.
proc buildFilecontents {filename {newFilename {}}} {
	set text [readFile $filename]
	# Fix end-of-line characters:
	regsub -all "\xa" $text "\xd" text
	set envName "filecontents"
	if { $newFilename == {} } {
		set envArg "{[file tail $filename]}"
	} else {
		set envArg "{$newFilename}"
	}
	return [buildEnvironment $envName $envArg "$text\r" "\r\r"]
}

#--------------------------------------------------------------------------
# Page Layout:
#--------------------------------------------------------------------------

proc maketitle {} {
	global searchNoisily
	set searchString {\\document(class|style)(\[.*\])?\{.*\}}
	set searchResult [search -s -n -f 1 -m 0 -i 1 -r 1 $searchString 0]
	if {[llength $searchResult] == 0} {
		if {$searchNoisily} {beep}
		message "can\'t find \\documentclass or \\documentstyle"
	} else {
		set searchPos [lindex $searchResult 1]
		set searchString {\\begin\{document\}}
		set searchResult [search -s -n -f 1 -m 0 -i 1 -r 1 $searchString $searchPos]
		if {[llength $searchResult] == 0} {
			if {$searchNoisily} {beep}
			message "can\'t find \\begin\{document\}"
		} else {
			goto [lindex $searchResult 1]
			set currentPos [getPos]
			set txt "\r\r% Definition of title page:"
			append txt "\r\\title\{"
			append txt "\r\t\r\}"
			append txt "\r\\author\{"
			append txt "\r\t\t% insert author(s) here"
			append txt "\r\}"
			append txt "\r\\date\{\}\t% optional"
			append txt "\r\r\\maketitle"
			elec::Insertion $txt
			message "insert title"
		}
	}
}
proc abstract {} { doWrapEnvironment "abstract" }
proc titlepage {} { doWrapEnvironment "titlepage" }
proc getPagestyle {} {
	catch {prompt "Choose a pagestyle:" "plain" "" "plain" "empty" \
				  "headings" "myheadings"} pagestyleName
	if {$pagestyleName != "cancel"} {
		return $pagestyleName
	} else {
		return ""
	}
}
proc pagestyle {} {
	set pagestyleName [getPagestyle]
	if {$pagestyleName != ""} {
		openingCarriageReturn
		insertObject "\\pagestyle\{$pagestyleName\}"
		closingCarriageReturn
	}
}
proc thispagestyle {} {
	set pagestyleName [getPagestyle]
	if {$pagestyleName != ""} {
		openingCarriageReturn
		insertObject "\\thispagestyle\{$pagestyleName\}"
		closingCarriageReturn
	}
}
proc getPagenumberingStyle {} {
	catch {prompt "Choose a pagenumbering style:" "arabic" "" "arabic" \
				  "roman" "Roman" "alph" "Alph"} pagenumberingStyle
	if {$pagenumberingStyle != "cancel"} {
		return $pagenumberingStyle
	} else {
		return ""
	}
}
proc pagenumbering {} {
	set pagenumberingStyle [getPagenumberingStyle]
	if {$pagenumberingStyle != ""} {
		openingCarriageReturn
		insertObject "\\pagenumbering\{$pagenumberingStyle\}"
		closingCarriageReturn
	}
}
proc twocolumn {} {
	openingCarriageReturn
	insertObject "\\twocolumn"
	closingCarriageReturn
}
proc onecolumn {} {
	openingCarriageReturn
	insertObject "\\onecolumn"
	closingCarriageReturn
}

#--------------------------------------------------------------------------
# Sectioning:
#--------------------------------------------------------------------------

proc sectioning {item} {
	append left [openingCarriageReturn] "\\${item}\{"
	append right "\}"
	if [regexp (part|chapter) $item] {
		append right "\\thispagestyle{empty}"
	}
	append right "" [closingCarriageReturn]
	if [wrapObject $left $right] {
		message "don't forget the label <Ctl Opt L>"
	} else {
		message "type the [string trim $item {*}] name and don't forget the label <Ctl Opt L>"
	}
}

proc appendix {} {
	insertObject "[openingCarriageReturn]\\appendix[closingCarriageReturn]"
}

#--------------------------------------------------------------------------
# Text Style:
#--------------------------------------------------------------------------

proc emph {} {
	if {[wrapObject "\\emph{" "}"]} {
		message "selected text has been emphasized"
	} else {
		message "enter text to be emphasized"
	}
}
proc textup {} {
	if {[wrapObject "\\textup{" "}"]} {
		message "selected text has upright shape"
	} else {
		message "enter text to have upright shape"
	}
}
proc textit {} {
	if {[wrapObject "\\textit{" "}"]} {
		message "selected text has italic shape"
	} else {
		message "enter text to have italic shape"
	}
}
proc textsl {} {
	if {[wrapObject "\\textsl{" "}"]} {
		message "selected text has slanted shape"
	} else {
		message "enter text to have slanted shape"
	}
}
proc textsc {} {
	if {[wrapObject "\\textsc{" "}"]} {
		message "selected text has small caps shape"
	} else {
		message "enter text to have small caps shape"
	}
}
proc textmd {} {
	if {[wrapObject "\\textmd{" "}"]} {
		message "selected text has been set in medium series"
	} else {
		message "enter text to be set in medium series"
	}
}
proc textbf {} {
	if {[wrapObject "\\textbf{" "}"]} {
		message "selected text has been set in bold series"
	} else {
		message "enter text to be set in bold series"
	}
}
proc textrm {} {
	if {[wrapObject "\\textrm{" "}"]} {
		message "selected text has been set with roman family"
	} else {
		message "enter text to be set using roman family"
	}
}
proc textsf {} {
	if {[wrapObject "\\textsf{" "}"]} {
		message "selected text has been set with sans serif family"
	} else {
		message "enter text to be set using sans serif family"
	}
}
proc texttt {} {
	if {[wrapObject "\\texttt{" "}"]} {
		message "selected text has been set with typewriter family"
	} else {
		message "enter text to be set using typewriter family"
	}
}
proc textnormal {} {
	if {[wrapObject "\\textnormal{" "}"]} {
		message "selected text has been set with normal style"
	} else {
		message "enter text to be set using normal style"
	}
}

proc em {} {
	if {[wrapObject "{\\em " "}"]} {
		message "emphasized text set"
	} else {
		message "enter text to be emphasized"
	}
}
proc upshape {} {
	if {[wrapObject "{\\upshape " "}"]} {
		message "text set in upright shape"
	} else {
		message "enter text to be set in upright shape"
	}
}
proc itshape {} {
	if {[wrapObject "{\\itshape " "}"]} {
		message "text set in italics shape"
	} else {
		message "enter text to be set in italics shape"
	}
}
proc slshape {} {
	if {[wrapObject "{\\slshape " "}"]} {
		message "text set in slanted shape"
	} else {
		message "enter text to be set in slanted shape"
	}
}
proc scshape {} {
	if {[wrapObject "{\\scshape " "}"]} {
		message "text set in small caps shape"
	} else {
		message "enter text to be set in small caps shape"
	}
}
proc mdseries {} {
	if {[wrapObject "{\\mdseries " "}"]} {
		message "text set in medium series"
	} else {
		message "enter text to be set in medium series"
	}
}
proc bfseries {} {
	if {[wrapObject "{\\bfseries " "}"]} {
		message "text set in bold series"
	} else {
		message "enter text to be set in bold series"
	}
}
proc rmfamily {} {
	if {[wrapObject "{\\rmfamily " "}"]} {
		message "text set in roman family"
	} else {
		message "enter text to be set in roman family"
	}
}
proc sffamily {} {
	if {[wrapObject "{\\sffamily " "}"]} {
		message "text set in sans serif family"
	} else {
		message "enter text to be set in sans serif family"
	}
}
proc ttfamily {} {
	if {[wrapObject "{\\ttfamily " "}"]} {
		message "text set in typewriter family"
	} else {
		message "enter text to be set in typewriter family"
	}
}
proc normalfont {} {
	if {[wrapObject "{\\normalfont " "}"]} {
		message "text set in normal style"
	} else {
		message "enter text to be set in normal style"
	}
}

#--------------------------------------------------------------------------
# Text Size:
#--------------------------------------------------------------------------

proc doTextSize {textSize} {
	if {[wrapObject "{\\$textSize " "}"]} {
		message "$textSize text set"
	} else {
		message "enter $textSize text"
	}
}

#--------------------------------------------------------------------------
# Text Commands:
#--------------------------------------------------------------------------

proc textsuperscript {} {
	if {[wrapObject "\\textsuperscript{" "}"]} {
		message "text superscripted"
	} else {
		message "enter superscripted text"
	}
}
proc textcircled {} {
	if {[wrapObject "\\textcircled{" "}"]} {
		message "text circled"
	} else {
		message "enter circled text"
	}
}

#--------------------------------------------------------------------------
# International:
#--------------------------------------------------------------------------

proc {} {} {
	if {[wrapObject "\\`{" "}"]} {
		message "accent set"
	} else {
		message "enter single character"
	}
}
proc {} {} {
	if {[wrapObject "\\'{" "}"]} {
		message "accent set"
	} else {
		message "enter single character"
	}
}
proc {} {} {
	if {[wrapObject "\\^{" "}"]} {
		message "accent set"
	} else {
		message "enter single character"
	}
}
proc {} {} {
	if {[wrapObject "\\\"{" "}"]} {
		message "accent set"
	} else {
		message "enter single character"
	}
}
proc {} {} {
	if {[wrapObject "\\~{" "}"]} {
		message "accent set"
	} else {
		message "enter single character"
	}
}
proc {} {} {insertObject "\\c\{c\}"}
proc {} {} {insertObject "\\c\{C\}"}
proc {} {} {insertObject "\\oe"}
proc {} {} {insertObject "\\OE"}
proc {} {} {insertObject "\\ae"}
proc {} {} {insertObject "\\AE"}
proc {} {} {insertObject "\\aa"}
proc {} {} {insertObject "\\AA"}
proc {} {} {insertObject "\\o"}
proc {} {} {insertObject "\\O"}
proc {ss} {} {insertObject "\\ss"}
proc {SS} {} {insertObject "\\SS"}
proc {} {} {insertObject "?`"}
proc {} {} {insertObject "!`"}

#--------------------------------------------------------------------------
# Environments:
#--------------------------------------------------------------------------

proc enumerate {} {
	global promptNoisily useStatusBar
	set envName "enumerate"
	if {$promptNoisily && $useStatusBar} {beep}
	catch {sPrompt "$envName:  how many items?" 3} numberItems
	if {$numberItems == "cancel"} {
		return
	} elseif {![isPositiveInteger $numberItems]} {
		beep
		message "invalid input:  unsigned, postive integer required"
		return
	}
	if {$numberItems} {
		set body "\t\\item  "
		for {set i 1} {$i < $numberItems} {incr i} {
			append body "\r\r\t\\item  "
		}
		append body "\r"
	} else {
		set body "\t\r"
	}
	if {[insertEnvironment $envName "" $body]} {
		message "type first item"
	}
}
proc itemize {} {
	global promptNoisily useStatusBar
	set envName "itemize"
	if {$promptNoisily && $useStatusBar} {beep}
	catch {sPrompt "$envName:  how many items?" 3} numberItems
	if {$numberItems == "cancel"} {
		return
	} elseif {![isPositiveInteger $numberItems]} {
		beep
		message "invalid input:  unsigned, postive integer required"
		return
	}
	if {$numberItems} {
		set body "\t\\item  "
		for {set i 1} {$i < $numberItems} {incr i} {
			append body "\r\r\t\\item  "
		}
		append body "\r"
	} else {
		set body "\t\r"
	}
	if {[insertEnvironment $envName "" $body]} {
		message "type first item"
	}
}
proc description {} {
	global promptNoisily useStatusBar
	set envName "description"
	if {$promptNoisily && $useStatusBar} {beep}
	catch {sPrompt "$envName:  how many items?" 3} numberItems
	if {$numberItems == "cancel"} {
		return
	} elseif {![isPositiveInteger $numberItems]} {
		beep
		message "invalid input:  unsigned, postive integer required"
		return
	}
	if {$numberItems} {
		set body "\t\\item\[\]  "
		for {set i 1} {$i < $numberItems} {incr i} {
			append body "\r\r\t\\item\[\]  "
		}
		append body "\r"
	} else {
		set body "\t\r"
	}
	if {[insertEnvironment $envName "" $body]} {
		message "type first item label"
	}
}
proc thebibliography {} {
	global promptNoisily useStatusBar
	set envName "thebibliography"
	if {$promptNoisily && $useStatusBar} {beep}
	catch {sPrompt "$envName:  how many items?" 3} numberItems
	if {$numberItems == "cancel"} {
		return
	} elseif {![isPositiveInteger $numberItems]} {
		beep
		message "invalid input:  unsigned, postive integer required"
		return
	}
	set arg "{9}"
	if {$numberItems} {
		if {$numberItems > 9} {set arg "{99}"}
		set body "\t\\bibitem{}  "
		for {set i 1} {$i < $numberItems} {incr i} {
			append body "\r\r\t\\bibitem{}  "
		}
		append body "\r"
	} else {
		set body "\t\r"
	}
	if {[insertEnvironment $envName $arg $body]} {
		set searchResult [search -s -n -f 1 -m 0 -i 1 -r 1 {(9)+} [getPos]]
		eval select $searchResult
		message "Change the length of the key field?"
	}
}
proc slide {} { doWrapEnvironment "slide" }
proc overlay {} { doWrapEnvironment "overlay" }
proc note {} { doWrapEnvironment "note" }
proc figure {} {
	global TeXmodeVars
	set envName "figure"
	set envArg "tbp"
	set arg "\[$envArg\]"
	set theIndentation [getIndentation [getPos]]
	append arg "\r$theIndentation\t\\centering"
	set body ""
	set macro [car $TeXmodeVars(boxMacroNames)]
	if { $macro != "" } {
		set restOfMacros [cdr $TeXmodeVars(boxMacroNames)]
		if { ![llength $restOfMacros] } {
			append body "\t\\$macro{}\r"
		} else {
			set cmd [list prompt "Choose a box macro:"]
			lappend cmd $macro "" 
			foreach boxMacroName $TeXmodeVars(boxMacroNames) {
				lappend cmd $boxMacroName
			}
			catch $cmd macro
			if {$macro == "cancel"} {
				message "operation canceled"
				return
			} elseif {$macro == ""} {
				# do nothing
			} else {
				append body "\t\\$macro{}\r"
			}
		}
	}
	append body "\t\\caption{}\r"
	append body "\t\\label{fig:}\r"
	if { $macro == "" } {
		wrapEnvironment $envName $arg $body
	} else {
		if {![insertEnvironment $envName $arg $body]} {return}
	}
	set searchResult [search -s -n -f 1 -m 0 -i 1 -r 1 "$envArg" [getPos]]
	eval select $searchResult
	message "Modify this argument?  (t=top; b=bottom; p=page; h=here; !=try harder)"
}
proc table {} {
	set envName "table"
	set envArg "tbp"
	set arg "\[$envArg\]"
	set theIndentation [getIndentation [getPos]]
	append arg "\r$theIndentation\t\\centering"
	# The following statement puts the caption at the top:
	append arg "\r$theIndentation\t\\caption{}"
	# The following statement puts the caption at the bottom:
# 	set body "\t\\caption{}\r"
	append body "\t\\label{tbl:}\r"
	wrapEnvironment $envName $arg $body
	set searchResult [search -s -n -f 1 -m 0 -i 1 -r 1 "$envArg" [getPos]]
	eval select $searchResult
	message "Modify this argument?  (t=top; b=bottom; p=page; h=here; !=try harder)"
}
proc buildRow {jmax} {
	set txt ""
	for {set j 1} {$j < $jmax} {incr j} {
		append txt " & "
	}
	return $txt
}
proc tabular {} {
	global promptNoisily useStatusBar
	set envName "tabular"
	if {$promptNoisily && $useStatusBar} {beep}
	catch {sPrompt "$envName:  how many rows?" 3} numberRows
	if {$numberRows == "cancel"} {
		return
	} elseif {![isPositiveInteger $numberRows]} {
		beep
		message "invalid input:  unsigned, postive integer required"
		return
	}
	if {$promptNoisily && $useStatusBar} {beep}
	catch {sPrompt "$envName:  how many columns?" 3} numberCols
	if {$numberCols == "cancel"} {
		return
	} elseif {![isPositiveInteger $numberCols]} {
		beep
		message "invalid input:  unsigned, postive integer required"
		return
	}
	set arg "{|"
	for {set j 1} {$j <= $numberCols} {incr j} {
		append arg "c|"
	}
	append arg "}"
	set body "\t\\hline\r"
	for {set i 1} {$i <= $numberRows} {incr i} {
		append body "\t[buildRow $numberCols]"
		append body "  \\\\\r\t\\hline\r"
	}
	if {[insertEnvironment $envName $arg $body]} {
		set searchResult [search -s -n -f 1 -m 0 -i 1 -r 1 {(c|\|)+} [getPos]]
		select [lindex $searchResult 0] [lindex $searchResult 1]
		message "Modify this argument?"
	}
}
proc verbatim {} { doWrapEnvironment "verbatim" }
proc quote {} { doWrapEnvironment "quote" }
proc quotation {} { doWrapEnvironment "quotation" }
proc verse {} { doWrapEnvironment "verse" }
proc flushleft {} { doWrapEnvironment "flushleft" }
proc center {} { doWrapEnvironment "center" }
proc flushright {} { doWrapEnvironment "flushright" }

#--------------------------------------------------------------------------
# Boxes:
#--------------------------------------------------------------------------

proc mbox {} {
	if {[wrapObject "\\mbox{" "}"]} {
		message "mbox set"
	} else {
		message "enter text"
	}
}
proc makebox {} {
	if {[wrapObject "\\makebox\[\]\[\]{" "}"]} {
		message "makebox set; enter the width and position"
	} else {
		message "enter the width and position of the makebox, then the text"
	}
}
proc fbox {} {
	if {[wrapObject "\\fbox{" "}"]} {
		message "fbox set"
	} else {
		message "enter text"
	}
}
proc framebox {} {
	if {[wrapObject "\\framebox\[\]\[\]{" "}"]} {
		message "framebox set; enter the width and position"
	} else {
		message "enter the width and position of the framebox, then the text"
	}
}
proc newsavebox {} {
	if {[wrapObject "\\newsavebox{" "}"]} {
		message "newsavebox defined"
	} else {
		message "enter the command name of the sbox or savebox"
	}
}
proc sbox {} {
	if {[wrapObject "\\sbox{}{" "}"]} {
		message "sbox set; enter the command name"
	} else {
		message "enter the command name of the sbox, then the text"
	}
}
proc savebox {} {
	if {[wrapObject "\\savebox{}\[\]\[\]{" "}"]} {
		message "savebox set; enter the command name"
	} else {
		message "enter the command name of the savebox"
	}
}
proc usebox {} {
	if {[wrapObject "\\usebox{" "}"]} {
		message "usebox declared"
	} else {
		message "enter the command name of the sbox or savebox"
	}
}
proc raisebox {} {
	if {[wrapObject "\\raisebox{}\[\]\[\]{" "}"]} {
		message "raisebox set; enter the displacement"
	} else {
		message "enter the displacement of the raisebox"
	}
}
proc parbox {} {
	if {[wrapObject "\\parbox\[\]\{\}{" "}"]} {
		message "parbox set; enter the position and width"
	} else {
		message "enter the position \[b|c|t\] and width of the parbox, then the text"
	}
}
proc minipage {} {
	set arg "\[\]{}"
	wrapEnvironment "minipage" $arg ""
	message "enter the position \[b|c|t\] of the minipage, then the width"
}
proc rule {} {
	insertObject "\\rule\[\]\{\}{}"
	message "enter the displacement of the rule, then width and height"
}

#--------------------------------------------------------------------------
# Misc:
#--------------------------------------------------------------------------

proc verb {} {
	if {[wrapObject "\\verb|" "|"]} {
		message "verbatim text set"
	} else {
		message "enter verbatim text"
	}
}
proc footnote {} {
	if {[wrapObject "\\footnote{" "}"]} {
		message "footnote set"
	} else {
		message "enter footnote"
	}
}
proc marginalNote {} {
	if {[wrapObject "\\marginpar{" "}"]} {
		message "marginal note set"
	} else {
		message "enter marginal note"
	}
}

proc insertLabel {} {
	if {[wrapObject "\\label{" "}"]} {
		message "label defined"
	} else {
		message "enter label"
	}
}
proc ref {} { 
	if {[wrapObject "\\ref{" "}" 1]} {
		message "reference made"
	} else {
		message "enter reference label"
	}
}
proc pageref {} { 
	if {[wrapObject "\\pageref{" "}" 1]} {
		message "page reference made"
	} else {
		message "enter page reference label"
	}
}
proc cite {} {
	if {[wrapObject "\\cite{" "}" 1]} {
		message "citation made"
	} else {
		message "enter citation key"
	}
}
proc nocite {} {
	if {[wrapObject "\\nocite{" "}"]} {
		message "citation added to the list"
	} else {
		message "enter citation key"
	}
}

# Insert an \item or a \bibitem, depending on the context.
proc insertItem {} {
	set command [eval getText [searchEnvironment]]
	set environment [extractCommandArg $command]
	switch $environment {
		"itemize" {
			set text "\\item  "
		}
		"enumerate" {
			set text "\\item  "
		}
		"description" {
			set text "\\item\[\]  "
		}
		"thebibliography" {
			set text "\\bibitem{}  "
		}
		default {
			beep
			message "insertItem: cursor in $environment environment"
			return
		}
	}
	set pos [getPos]
	# Indentation should mirror that of an existing \item
	# (if it exists)
	elec::Insertion [openingCarriageReturn]$text
}

proc quotes {} {
	if {[wrapObject "`" "'"]} {
		message "text quoted"
	} else {
		message "enter text"
	}
}
proc dblQuotes {} {
	if {[wrapObject "``" "''"]} {
		message "text double quoted"
	} else {
		message "enter text"
	}
}

proc texLogo {} {insertObject "\\TeX"}
proc latexLogo {} {insertObject "\\LaTeX"}
proc latex2eLogo {} {insertObject "\\LaTeXe"}
proc today {} {insertObject "\\today"}

proc dag {} {insertObject "\\dag"}
proc ddag {} {insertObject "\\ddag"}
proc sectionMark {} {insertObject "\\S"}
proc paragraphMark {} {insertObject "\\P"}
proc copyright {} {insertObject "\\copyright"}
proc pounds {} {insertObject "\\pounds"}


#############################################################################
# 
# Math Mode Macros
#
#############################################################################

#--------------------------------------------------------------------------
# Math Modes:
#--------------------------------------------------------------------------

proc texMath {} {
	checkMathMode "texMath" 0
	if {[wrapObject "$" "$"]} {
		message "formula set"
	} else {
		message "enter formula"
	}
}
proc texDisplaymath {} {
	checkMathMode "texDisplaymath" 0
	if {[wrapObject "$$" "$$"]} {
		message "displayed formula set"
	} else {
		message "enter displayed formula"
	}
}
proc latexMath {} {
	checkMathMode "latexMath" 0
	if {[wrapObject "\\( " " \\)"]} {
		message "formula set"
	} else {
		message "enter formula"
	}
}
proc latexDisplaymath {} {
	checkMathMode "latexDisplaymath" 0
	if {[wrapObject "\\\[ " " \\\]"]} {
		message "displayed formula set"
	} else {
		message "enter displayed formula"
	}
}

#--------------------------------------------------------------------------
# Math Style:
#--------------------------------------------------------------------------

proc doMathStyle {mathStyle description} {
	checkMathMode "$mathStyle" 1
	if {[wrapObject "\\$mathStyle{" "}"]} {
		message "selected text is $description"
	} else {
		message "enter text to be $description"
	}
}
proc doUppercaseMathStyle {mathStyle description} {
	checkMathMode "$mathStyle" 1
	# Allow upper-case alphabetic arguments only:
	if {[isSelection] && (![isUppercase] || ![isAlphabetic])} {
		beep
		alertnote "argument to \\$mathStyle must be UPPERCASE alphabetic"
		return
	}
	if {[wrapObject "\\$mathStyle{" "}"]} {
		message "selected text is $description"
	} else {
		message "enter text to be $description (UPPERCASE letters only)"
	}
}

#--------------------------------------------------------------------------
# Math Environments:
#--------------------------------------------------------------------------

proc math {} { checkMathMode "math" 0; doWrapEnvironment "math" }
proc displaymath {} {
	checkMathMode "displaymath" 0
	global TeXmodeVars
	if { $TeXmodeVars(useBrackets) } {
		doWrapStructure {\[} {} {\]}
	} else {
		doWrapEnvironment "displaymath"
	}
}
proc equation {} {
	checkMathMode "equation" 0
	set envName "equation"
	set body "\t\\label{eq:}\r"
	if {[wrapEnvironment $envName "" $body]} {
		set msgText "equation wrapped"
	} else {
		set msgText "enter equation"
	}
	message $msgText
}
proc eqnarrayStar {} {
	global promptNoisily useStatusBar
	checkMathMode "eqnarrayStar" 0
	set envName "eqnarray*"
	if {$promptNoisily && $useStatusBar} {beep}
	catch {sPrompt "$envName:  how many rows?" 3} numberRows
	if {$numberRows == "cancel"} {
		return
	} elseif {![isPositiveInteger $numberRows]} {
		beep
		message "invalid input:  unsigned, postive integer required"
		return
	}
	set row "\t[buildRow 3]"
	for {set i 1} {$i < $numberRows} {incr i} {
		append body $row
		append body "  \\\\\r"
	}
	append body $row
	append body "\r"
	if {[insertEnvironment $envName "" $body]} {
		message "type first item"
	}
}
proc eqnarray {} {
	global promptNoisily useStatusBar
	checkMathMode "eqnarray" 0
	set envName "eqnarray"
	if {$promptNoisily && $useStatusBar} {beep}
	catch {sPrompt "$envName:  how many rows?" 3} numberRows
	if {$numberRows == "cancel"} {
		return
	} elseif {![isPositiveInteger $numberRows]} {
		beep
		message "invalid input:  unsigned, postive integer required"
		return
	}
	set row "\t[buildRow 3]\r\t\\label{eq:}"
	for {set i 1} {$i < $numberRows} {incr i} {
		append body $row
		append body "  \\\\\r"
	}
	append body $row
	append body "\r"
	if {[insertEnvironment $envName "" $body]} {
		message "type first item"
	}
}
proc myArray {} {
	global promptNoisily useStatusBar
	checkMathMode "myArray" 1
	set envName "array"
	if {$promptNoisily && $useStatusBar} {beep}
	catch {sPrompt "$envName:  how many rows?" 3} numberRows
	if {$numberRows == "cancel"} {
		return
	} elseif {![isPositiveInteger $numberRows]} {
		beep
		message "invalid input:  unsigned, postive integer required"
		return
	}
	if {$promptNoisily && $useStatusBar} {beep}
	catch {sPrompt "$envName:  how many columns?" 3} numberCols
	if {$numberCols == "cancel"} {
		return
	} elseif {![isPositiveInteger $numberCols]} {
		beep
		message "invalid input:  unsigned, postive integer required"
		return
	}
	set arg "{"
	for {set j 1} {$j <= $numberCols} {incr j} {
		append arg "c"
	}
	append arg "}"
	set row "\t[buildRow $numberCols]"
	for {set i 1} {$i < $numberRows} {incr i} {
		append body $row
		append body "  \\\\\r"
	}
	append body $row
	append body "\r"
	if {[insertEnvironment $envName $arg $body]} {
		set searchResult [search -s -n -f 1 -m 0 -i 1 -r 1 {c+} [getPos]]
		select [lindex $searchResult 0] [lindex $searchResult 1]
		message "Modify this argument?  (c=center; l=left; r=right; p{width}; a{text})"
	}
}

#--------------------------------------------------------------------------
# Formulas:
#--------------------------------------------------------------------------

proc subscript {} {
	checkMathMode "subscript" 1
	if {[wrapObject "_{" "}"]} {
		message "subscript set"
	} else {
		message "enter subscript"
	}
}
proc superscript {} {
	checkMathMode "superscript" 1
	if {[wrapObject "^{" "}"]} {
		message "superscript set"
	} else {
		message "enter superscript"
	}
}
proc fraction {} {
	checkMathMode "fraction" 1
	set currentPos [getPos]
	if {[isSelection]} {
		set selection [getSelect]
		set args [split $selection /]
		set len [llength $args]
		deleteText $currentPos [selEnd]
		if {$len == 1} {
			insertObject "\\frac{$selection}{}"
			message "enter denominator"
		} else {
			set firstArg [lindex $args 0]
			set restArgs [lrange $args 1 [expr $len-1]]
			insertObject "\\frac{$firstArg}{[join $restArgs /]}"
			if {$len > 2} {message "beware of multiple /"}
		}
	} else {
		insertObject "\\frac{}{}"
		message "enter numerator"
	}
}
proc squareRoot {} {
	checkMathMode "squareRoot" 1
	if {[wrapObject "\\sqrt{" "}"]} {
		message "square root set"
	} else {
		message "enter formula"
	}
}
proc nthRoot {} {
	checkMathMode "nthRoot" 1
	if {[wrapObject "\\sqrt\[\]{" "}"]} {
		message "enter root"
	} else {
		message "enter root, then formula"
	}
}
proc oneParameter {} {
	checkMathMode "oneParameter" 1
	if {[wrapObject "\\{" "}"]} {
		message "enter command name"
	} else {
		message "enter command name, press <Tab>, enter argument"
	}
}
proc twoParameters {} {
	checkMathMode "twoParameters" 1
	if {[wrapObject "\\{" "}{}"]} {
		message "enter command name"
	} else {
		message "enter command name, press <Tab>, enter argument, etc."
	}
}

#--------------------------------------------------------------------------
# Large Ops:
#--------------------------------------------------------------------------

proc insertLargeOp {commandName} {
	checkMathMode "$commandName" 1
	insertObject "\\$commandName\_{}^{}"
}
proc sum {} {insertLargeOp "sum"}
proc prod {} {insertLargeOp "prod"}
proc coprod {} {insertLargeOp "coprod"}
proc int {} {insertLargeOp "int"}
proc oint {} {insertLargeOp "oint"}
proc bigcap {} {insertLargeOp "bigcap"}
proc bigcup {} {insertLargeOp "bigcup"}
proc bigsqcup {} {insertLargeOp "bigsqcup"}
proc bigvee {} {insertLargeOp "bigvee"}
proc bigwedge {} {insertLargeOp "bigwedge"}
proc bigodot {} {insertLargeOp "bigodot"}
proc bigotimes {} {insertLargeOp "bigotimes"}
proc bigoplus {} {insertLargeOp "bigoplus"}
proc biguplus {} {insertLargeOp "biguplus"}

#--------------------------------------------------------------------------
# Delimiters:
#--------------------------------------------------------------------------

proc delimitObject {leftDelim rightDelim} {
	if {[wrapObject $leftDelim $rightDelim]} {
		message "formula delimited"
	} else {
		message "enter formula"
	}
}
proc parentheses {} { checkMathMode "parentheses" 1; delimitObject "(" ")" }
proc brackets {} { checkMathMode "brackets" 1; delimitObject "\[" "\]" }
proc braces {} { checkMathMode "braces" 1; delimitObject "\\\{" "\\\}" }
proc absoluteValue {} { checkMathMode "absoluteValue" 1; delimitObject "|" "|" }
proc getDelims {} {
	catch {prompt "Choose delimiters:" "parentheses" "" "parentheses" \
				  "brackets" "braces" "angle brackets" "vertical bars" \
				  "double bars" "ceiling" "floor"} delimType
	if {$delimType != "cancel"} {
		switch $delimType {
			"parentheses" {
				set leftDelim "("
				set rightDelim ")"
			}
			"brackets" {
				set leftDelim "\["
				set rightDelim "\]"
			}
			"braces" {
				set leftDelim "\\\{"
				set rightDelim "\\\}"
			}
			"vertical bars" {
				set leftDelim "|"
				set rightDelim "|"
			}
			"double bars" {
				set leftDelim "\\|"
				set rightDelim "\\|"
			}
			"angle brackets" {
				set leftDelim "\\langle"
				set rightDelim "\\rangle"
			}
			"ceiling" {
				set leftDelim "\\lceil"
				set rightDelim "\\rceil"
			}
			"floor" {
				set leftDelim "\\lfloor"
				set rightDelim "\\rfloor"
			}
			default {
				alertnote "\"$delimType\" not recognized"
				return ""
			}
		}
		return [list $leftDelim $rightDelim]
	} else {return ""}
}
proc otherDelims {} {
	checkMathMode "otherDelims" 1
	set delims [getDelims]
	if {$delims != ""} {
		set leftDelim [lindex $delims 0]
		set rightDelim [lindex $delims 1]
		delimitObject "$leftDelim" "$rightDelim"
	}
}
proc {half-openInterval} {} {
	checkMathMode "half-openInterval" 1; delimitObject "(" "\]"
}
proc {half-closedInterval} {} {
	checkMathMode "half-closedInterval" 1; delimitObject "\[" ")"
}

proc insertBigDelims {leftDelim rightDelim isMultiline} {
	checkMathMode "insertBigDelims" 1
	if {$isMultiline} {
		doWrapStructure $leftDelim "" $rightDelim
	} else {
		if { [wrapObject $leftDelim $rightDelim] } {
			message "formula delimited"
		} else {
			message "enter formula"
		}
	}
}
proc bigParens {} {
	checkMathMode "bigParens" 1; insertBigDelims "\\left(" "\\right)" 0
}
proc multiBigParens {} {
	checkMathMode "multiBigParens" 1; insertBigDelims "\\left(" "\\right)" 1
}
proc bigBrackets {} {
	checkMathMode "bigBrackets" 1; insertBigDelims "\\left\[" "\\right\]" 0
}
proc multiBigBrackets {} {
	checkMathMode "multiBigBrackets" 1; insertBigDelims "\\left\[" "\\right\]" 1
}
proc bigBraces {} {
	checkMathMode "bigBraces" 1; insertBigDelims "\\left\\\{" "\\right\\\}" 0
}
proc multiBigBraces {} {
	checkMathMode "multiBigBraces" 1; insertBigDelims "\\left\\\{" "\\right\\\}" 1
}
proc bigAbsValue {} {
	checkMathMode "bigAbsValue" 1; insertBigDelims "\\left|" "\\right|" 0
}
proc multiBigAbsValue {} {
	checkMathMode "multiBigAbsValue" 1; insertBigDelims "\\left|" "\\right|" 1
}
proc doOtherBigDelims {name isMultiline} {
	checkMathMode $name 1
	set delims [getDelims]
	if {$delims != ""} {
		append leftDelim "\\left" [lindex $delims 0]
		append rightDelim "\\right" [lindex $delims 1]
		insertBigDelims "$leftDelim" "$rightDelim" $isMultiline
	}
}
proc otherBigDelims {} {
	doOtherBigDelims "otherBigDelims" 0
}
proc otherMultiBigDelims {} {
	doOtherBigDelims "otherMultiBigDelims" 1
}
proc bigLeftBrace {} {
	checkMathMode "bigLeftBrace" 1
	insertBigDelims "\\left\\\{" "\\right." 0
}
proc multiBigLeftBrace {} {
	checkMathMode "multiBigLeftBrace" 1
	insertBigDelims "\\left\\\{" "\\right." 1
}
proc doOtherMixedBigDelims {name isMultiline} {
	checkMathMode $name 1
	catch {prompt "Choose LEFT delimiter:" "parenthesis" "" "parenthesis" \
	              "bracket" "brace" "vertical bar" "double bar" \
	              "angle bracket" "ceiling" "floor" "slash" "backslash" \
	              "none"} delimType
	if {$delimType != "cancel"} {
		switch $delimType {
			"parenthesis" {set leftDelim "("}
			"bracket" {set leftDelim "\["}
			"brace" {set leftDelim "\\\{"}
			"vertical bar" {set leftDelim "|"}
			"double bar" {set leftDelim "\\|"}
			"angle bracket" {set leftDelim "\\langle"}
			"ceiling" {set leftDelim "\\lceil"}
			"floor" {set leftDelim "\\lfloor"}
			"slash" {set leftDelim "/"}
			"backslash" {set leftDelim "\\backslash"}
			"none" {set leftDelim "."}
			default {
				alertnote "\"$delimType\" not recognized"
				return
			}
		}
		catch {prompt "Choose RIGHT delimiter:" "parenthesis" "" "parenthesis" \
		              "bracket" "brace" "vertical bar" "double bar" \
		              "angle bracket" "ceiling" "floor" "slash" "backslash" \
		              "none"} delimType
		if {$delimType != "cancel"} {
			switch $delimType {
				"parenthesis" {set rightDelim ")"}
				"bracket" {set rightDelim "\]"}
				"brace" {set rightDelim "\\\}"}
				"vertical bar" {set rightDelim "|"}
				"double bar" {set rightDelim "\\|"}
				"angle bracket" {set rightDelim "\\rangle"}
				"ceiling" {set rightDelim "\\rceil"}
				"floor" {set rightDelim "\\rfloor"}
				"slash" {set rightDelim "/"}
				"backslash" {set rightDelim "\\backslash"}
				"none" {set rightDelim "."}
				default {
					alertnote "\"$delimType\" not recognized"
					return
				}
			}
			insertBigDelims "\\left$leftDelim" "\\right$rightDelim" $isMultiline
		}
	}
}
proc otherMixedBigDelims {} {
	doOtherMixedBigDelims "otherMixedBigDelims" 0
}
proc otherMultiMixedBigDelims {} {
	doOtherMixedBigDelims "otherMultiMixedBigDelims" 1
}

#--------------------------------------------------------------------------
# Accents:
#--------------------------------------------------------------------------

proc acute {} {
	checkMathMode "acute" 1
	if {[isSelection] > 1} {
		alertnote "Warning: only a single character may be accented!"
	}
	if {[wrapObject "\\acute{" "}"]} {
		message "accent set"
	} else {
		message "enter one character"
	}
}
proc bar {} {
	checkMathMode "bar" 1
	if {[isSelection] > 1} {
		alertnote "Warning: only a single character may be accented!"
	}
	if {[wrapObject "\\bar{" "}"]} {
		message "accent set"
	} else {
		message "enter one character"
	}
}
proc breve {} {
	checkMathMode "breve" 1
	if {[isSelection] > 1} {
		alertnote "Warning: only a single character may be accented!"
	}
	if {[wrapObject "\\breve{" "}"]} {
		message "accent set"
	} else {
		message "enter one character"
	}
}
proc check {} {
	checkMathMode "check" 1
	if {[isSelection] > 1} {
		alertnote "Warning: only a single character may be accented!"
	}
	if {[wrapObject "\\check{" "}"]} {
		message "accent set"
	} else {
		message "enter one character"
	}
}
proc dot {} {
	checkMathMode "dot" 1
	if {[isSelection] > 1} {
		alertnote "Warning: only a single character may be accented!"
	}
	if {[wrapObject "\\dot{" "}"]} {
		message "accent set"
	} else {
		message "enter one character"
	}
}
proc ddot {} {
	checkMathMode "ddot" 1
	if {[isSelection] > 1} {
		alertnote "Warning: only a single character may be accented!"
	}
	if {[wrapObject "\\ddot{" "}"]} {
		message "accent set"
	} else {
		message "enter one character"
	}
}
proc grave {} {
	checkMathMode "grave" 1
	if {[isSelection] > 1} {
		alertnote "Warning: only a single character may be accented!"
	}
	if {[wrapObject "\\grave{" "}"]} {
		message "accent set"
	} else {
		message "enter one character"
	}
}
proc hat {} {
	checkMathMode "hat" 1
	if {[isSelection] > 1} {
		alertnote "Warning: only a single character may be accented!"
	}
	if {[wrapObject "\\hat{" "}"]} {
		message "accent set"
	} else {
		message "enter one character"
	}
}
proc tilde {} {
	checkMathMode "tilde" 1
	if {[isSelection] > 1} {
		alertnote "Warning: only a single character may be accented!"
	}
	if {[wrapObject "\\tilde{" "}"]} {
		message "accent set"
	} else {
		message "enter one character"
	}
}
proc vec {} {
	checkMathMode "vec" 1
	if {[isSelection] > 1} {
		alertnote "Warning: only a single character may be accented!"
	}
	if {[wrapObject "\\vec{" "}"]} {
		message "accent set"
	} else {
		message "enter one character"
	}
}

proc widehat {} {
	checkMathMode "widehat" 1
	if {[isSelection] > 3} {
		alertnote "Warning: only a few characters may be accented!"
	}
	if {[wrapObject "\\widehat{" "}"]} {
		message "accent set"
	} else {
		message "enter a few characters"
	}
}
proc widetilde {} {
	checkMathMode "widetilde" 1
	if {[isSelection] > 3} {
		alertnote "Warning: only a few characters may be accented!"
	}
	if {[wrapObject "\\widetilde{" "}"]} {
		message "accent set"
	} else {
		message "enter a few characters"
	}
}


#--------------------------------------------------------------------------
# Grouping:
#--------------------------------------------------------------------------

proc underline {} {
	checkMathMode "underline" 1
	if {[wrapObject "\\underline{" "}"]} {
		message "selection underlined"
	} else {
		message "enter text"
	}
}
proc overline {} {
	checkMathMode "overline" 1
	if {[wrapObject "\\overline{" "}"]} {
		message "selection overlined"
	} else {
		message "enter text"
	}
}
proc underbrace {} {
	checkMathMode "underbrace" 1
	if {[wrapObject "\\underbrace{" "}"]} {
		message "selection underbraced"
	} else {
		message "enter text"
	}
}
proc overbrace {} {
	checkMathMode "overbrace" 1
	if {[wrapObject "\\overbrace{" "}"]} {
		message "selection overbraced"
	} else {
		message "enter text"
	}
}
proc overrightarrow {} {
	checkMathMode "overrightarrow" 1
	if {[wrapObject "\\overrightarrow{" "}"]} {
		message "selection overrightarrowed"
	} else {
		message "enter text"
	}
}
proc overleftarrow {} {
	checkMathMode "overleftarrow" 1
	if {[wrapObject "\\overleftarrow{" "}"]} {
		message "selection overleftarrowed"
	} else {
		message "enter text"
	}
}
proc stackrel {} {
	checkMathMode "stackrel" 1
	set currentPos [getPos]
	if {[insertObject "\\stackrel{}{}"]} {
		message "1st arg scriptstyle"
	}
}

#--------------------------------------------------------------------------
# Spacing:
#--------------------------------------------------------------------------

proc negThin {} {checkMathMode "negThin" 1; insertObject "\\!"}
proc thin {} {checkMathMode "thin" 1; insertObject "\\,"}
proc medium {} {checkMathMode "medium" 1; insertObject "\\:"}
proc thick {} {checkMathMode "thick" 1; insertObject "\\;"}
proc quad {} {checkMathMode "quad" 1; insertObject "\\quad"}
proc qquad {} {checkMathMode "qquad" 1; insertObject "\\qquad"}
proc hspace {} {
	checkMathMode "hspace" 1
	if {[wrapObject "\\hspace{" "}"]} {
		message "spacing set"
	} else {
		message "enter the desired horizontal spacing"
	}
}
proc vspace {} {
	checkMathMode "vspace" 1
	if {[wrapObject "\\vspace{" "}"]} {
		message "spacing set"
	} else {
		message "enter the desired horizontal spacing"
	}
}
proc hfill {} {checkMathMode "hfill" 1; insertObject "\\hfill"}
proc vfill {} {checkMathMode "vfill" 1; insertObject "\\vfill"}
proc smallskip {} {checkMathMode "smallskip" 1; insertObject "\\smallskip"}
proc medskip {} {checkMathMode "medskip" 1; insertObject "\\medskip"}
proc bigskip {} {checkMathMode "bigskip" 1; insertObject "\\bigskip"}

