#############################################################################
#############################################################################
#
# latexUtilities.tcl (sourced on demand)
#
#############################################################################
#
# Author:  Tom Scavo <trscavo@syr.edu>
#
#############################################################################
#############################################################################

# Delete all unnecessary comments from the current document:
proc deleteComments {} {
	switch [askyesno "Warning!  This operation can not be undone.  \
	  Continue anyway?"] {
		"yes" {}
		"no" { return }
	}
	replaceString {}
	eval select [search -f 1 -r 1 -i 1 -m 0 -- {^[ \t]*%.*\r} 0]
	replaceAll
	
	replaceString {}
	eval select [search -f 1 -r 1 -i 1 -m 0 -- {[ \t]+%.*} 0]
	replaceAll
	
	replaceString {\1%}
	eval select [search -f 1 -r 1 -i 1 -m 0 -- {([^\\](\\\\)*)%.*} 0]
	replaceAll
}

# Converts all straight quotes to their TeX equivalents.
proc convertQuotes {} {
	global searchNoisily
	message "working"
	watchCursor
	set messageString "selection"
	if {[set start [getPos]] == [set end [selEnd]]} {
		set messageString "document"
		set start 0
		set end [maxPos]
	}
	set text [getText $start $end]
	# Convert all left double quotes:
	set convert1 [regsub -all "\(\^\|\[\ \r\t\(\[\{\]\)\"" $text {\1``} text]
	# Convert all right double quotes:
	set convert2 [regsub -all "\(\[\^\\\\\]\)\"" $text {\1''} text]
	# Convert all left single quotes:
	set convert3 [regsub -all "\(\^\|\[\ \r\t\(\[\{\]\)\'" $text {\1`} text]
	if {$convert1 || $convert2 || $convert3} {
		replaceText $start $end $text
		message "all quotes in $messageString converted"
	} else {
		if {$searchNoisily} {beep}
		message "no quotes found in $messageString"
	}
}

# Convert all dollar signs to their LaTeX equivalents:
proc convertDollarSigns {} {
	global searchNoisily
	if {[isSelection]} {
		set messageString "selection"
	} else {
		set messageString "document"
	}
	set subs2 [convertDoubleDollarSigns]
	if {$subs2 == -1} {
		beep
		alertnote "unmatched double dollar signs in $messageString"
	} else {
		set subs1 [convertSingleDollarSigns]
		if {$subs1 == -1} {
			beep
			alertnote "unmatched single dollar sign in $messageString"
		} elseif {$subs1 == 0 && $subs2 == 0} {
			if {$searchNoisily} {beep}
			message "no dollar signs found in $messageString"
		} else {
			message "$subs1 pairs of \$\$ and $subs2 pairs of \$\$\$\$ removed from $messageString"
		}
	}
}

# Converts all $$...$$ pairs to \[...\] and returns the number of such 
# pairs converted.  If the dollar signs are unbalanced, does nothing 
# and returns -1.
proc convertDoubleDollarSigns {} {
	watchCursor
	set messageString "selection"
	if {[set start [getPos]] == [set end [selEnd]]} {
		set messageString "document"
		set start 0
		set end [maxPos]
	}
	set text [getText $start $end]
	set subs [regsub -all {(^|[^\\])\$\$([^$]*)\$\$} $text {\1\\[\2\\]} text]
	if {[containsDoubleDollarSigns $text]} {return -1}
	if {$subs} {
		replaceText $start $end $text
	}
	return [expr $subs]
}

# Returns true if the argument contains non-literal double dollar
# signs, and false otherwise.
proc containsDoubleDollarSigns {text} {
	return [regexp {(^|[^\\])\$\$} $text]
}

# Converts all $...$ pairs to \(...\), maintains the cursor position, 
# and returns the number of such pairs converted.  If the dollar signs 
# are unbalanced, does nothing and returns -1.
proc convertSingleDollarSigns {} {
	watchCursor
	set subs1 0; set subs2 0; set subs3 0
	set pos [getPos]; set pos2 $pos
	if {[set start $pos] == [set end [selEnd]]} {
		set isSelection 0
		set start 0
		set end [maxPos]
		set text1 [getText $start $pos]
		set subs1 [regsub -all {(^|[^\\])\$([^$]*)\$} $text1 {\1\\(\2\\)} text1]
		# Is there a dollar sign left over?  If so, search backward for this
		# dollar sign and prepare to do a substitution on the text to the right
		# of this dollar sign.
		if {[containsSingleDollarSign $text1]} {
			set searchString {[^\\]\$}
			set searchResult [search -s -n -f 0 -m 0 -i 1 -r 1 $searchString [expr $pos-1]]
			set pos2 [lindex $searchResult 0]
			set text1 [string range $text1 0 [expr $pos2 + (2 * $subs1)]]
			set pos [expr $pos + 2]
		}
		set text2 [getText $pos2 $end]
		set subs2 [regsub -all {(^|[^\\])\$([^$]*)\$} $text2 {\1\\(\2\\)} text2]
		# Is there a dollar sign left over?  If so, it's unbalanced.
		if {[containsSingleDollarSign $text2]} {return -1}
		append text $text1 $text2
	} else {
		set isSelection 1
		set text [getText $start $end]
		set subs3 [regsub -all {(^|[^\\])\$([^$]*)\$} $text {\1\\(\2\\)} text]
		# Is there a dollar sign left over?  If so, it's unbalanced.
		if {[containsSingleDollarSign $text]} {return -1}
	}
	if {$subs1 || $subs2 || $subs3} {
		replaceText $start $end $text
		# If there is a selection, just put it back.  Otherwise, adjust the
		# cursor position based on the number of substitutions.
		if {$isSelection} {
			set end [getPos]
			select $start $end
		} else {
			goto [expr $pos + (2 * $subs1)]
		}
	}
	return [expr $subs1 + $subs2 + $subs3]
}

# Returns true if the argument contains a non-literal dollar sign,
# and false otherwise.
proc containsSingleDollarSign {text} {
	return [regexp {(^|[^\\])\$} $text]
}

######################################################################
# 
# File:         accents.tcl
# Version:      1.01
# Purpose:      Translates Mac characters to their LaTeX equivalents 
#               and vice-versa
# Last Change:  1.5.97
# Language:     TCL
# Author:       F. Miguel Dionsio
# Email:        fmd@math.ist.utl.pt
# Copying:      Free
# 
# The purpose of this tcl script is to translate mac characters to 
# their LaTeX equivalents and vice-versa (from within the Mac 
# editor Alpha).  The characters are those of the Courier font, 
# with the exception of the ones that, in LaTeX, belong to math 
# mode (see list below).  This is useful to  people who share 
# their LaTeX sources with non-Mac users.
# 
# Two procedures are defined: accentsToLatexInRegion and 
# latexToAccentsInRegion that convert all SELECTED text.  The 
# conversion latexToAccentsInRegion tries to match (using regular 
# expressions) various forms of the character definition in LaTeX.
#
######################################################################
#
# List of characters (use Courier to see them):
# 
#		""	""	""	""	""	 ""
#		""	""	""	""	""	 ""
#		""	""	""	""
#		""	""	""	""
#		""	""	""	""
#		""	""	""	""
#		""	""	""	""	""
#		""	""	""	""	""
#		""	""	""	""
#		""	""	""	""
#		""	""	""	""
#		""	""	""	""
#		""	""	""
#		""	""	""	""	""	""	""	""
#		""	""	""
#		""	""
# 
#  WARNING:  appears different in Monaco
# 
#  List of OMITTED characters from Courier:
# 
#  ,,, , , and the following symbols from mathematical mode:
#  ,,,,,,,,,,,,,,,,,,,,,,
# 
# If you find some more problems please tell me. (fmd@math.ist.utl.pt)
#
######################################################################
# 
# Alpha commands used: getText, getPos, selEnd, replaceText; 
# all others are TCL built-ins
# 
######################################################################

######################################################################
#
# Returns, for argument "a" the regular expression 
# [ \t]*(a|{[ \t]*a[ \t]*}),
# used to look for alternative ways of writting, for example :
# \`a, \` a, \`{a}, etc.
#
######################################################################

proc rexp  {c} {
	set ws {[ \t]*}
	set reg $ws\($c|{$ws$c$ws}\)
	return "$reg"
}

######################################################################
# 
# This does the rest: defines the list of chars (all), the list of their 
# LaTex equivalent (texall) and the list of corresponding regular 
# expressions (regall).  When translating to LaTeX replaces all 
# ocurrences of each char by the corresponding LaTeX equivalent (using 
# regsub, see the TCL manual).  In the other direction replaces the text 
# that matches the regular expression by the corresponding char.
# 
######################################################################

proc accentsToLatexInRegion {} {replaceAccents 1}
proc latexToAccentsInRegion {} {replaceAccents 0}

proc  replaceAccents {ww} {
	if { [isSelection] } {
		set position [getPos]
		set endselection [selEnd]
	} else {
		switch [askyesno "Convert the entire document?"] {
			"yes" {
				set position 0
				set endselection [maxPos]
			}
			"no" { return }
		}
	}
	set text [getText $position $endselection]
	
	############################################################
	# 
	# List of characters to be replaced by LaTeX equivalent
	# 
	############################################################

	set as    [list "" "" "" "" ""  ""] 
	set cas   [list "" "" "" "" ""  ""] 
	set es    [list "" "" "" ""]
	set ces   [list "" "" "" ""]
	set is    [list "" "" "" ""]
	set cis   [list "" "" "" ""]
	set os    [list "" "" "" "" ""]
	set cos   [list "" "" "" "" ""]
	set us    [list "" "" "" ""]
	set cus   [list "" "" "" ""]
	set lig   [list "" "" "" ""]
	set oth1  [list "" "" "" ""]
	set coth1 [list "" "" ""]
	set oth2  [list "" "" "" "" "" "" "" ""]
	set oth3  [list "" "" ""]
	set oth4  [list "" ""]
	
	set all [concat $as $cas $es $ces $is $cis $os $cos $us $cus $lig $oth1 $coth1 $oth2 $oth3 $oth4]

	############################################################
	# 
	# List of the LaTex equivalents
	# 
	############################################################

	set quote {"}  
	set seplater {\\\\sepsep//}
      
	if {$ww == "1"} {      
		set texas    [list  "\\`{a}" "\\'{a}" "\\^{a}" "\\~{a}" "\\$quote{a}" "\\aa$seplater"] 
		set texcas   [list  "\\`{A}" "\\'{A}" "\\^{A}" "\\~{A}" "\\$quote{A}" "\\AA$seplater"]                                     
		set texes    [list  "\\`{e}" "\\'{e}" "\\^{e}" "\\$quote{e}"]                
		set texces   [list  "\\`{E}" "\\'{E}" "\\^{E}" "\\$quote{E}"]   
		set texis    [list  "\\`{\\i}" "\\'{\\i}" "\\^{\\i}"  "\\$quote{\\i}"]
		set texcis   [list  "\\`{I}"   "\\'{I}"   "\\^{I}" "\\$quote{I}"]
		set texos    [list  "\\`{o}" "\\'{o}" "\\^{o}" "\\~{o}" "\\$quote{o}"]
		set texcos   [list  "\\`{O}" "\\'{O}" "\\^{O}" "\\~{O}" "\\$quote{O}"]
		set texus    [list  "\\`{u}" "\\'{u}" "\\^{u}" "\\$quote{u}"]
		set texcus   [list  "\\`{U}" "\\'{U}" "\\^{U}" "\\$quote{U}"]
		set texlig   [list  "\\oe$seplater"  "\\OE$seplater" "\\ae$seplater" "\\AE$seplater"]
		set texoth1  [list  "\\c{c}" "\\~{n}" "\\$quote{y}" "\\dots$seplater"]
		set texcoth1 [list  "\\c{C}" "\\~{N}" "\\$quote{Y}" ]
		set texoth2  [list  "\\ss$seplater"  "?`" "!`" "\\pounds$seplater" "\\o$seplater" "\\O$seplater" "\\S$seplater" "\\P$seplater"]
		set texoth3  [list   "{\\leavevmode\\raise.585ex\\hbox{\\b{a}}}"  "{\\leavevmode\\raise.6ex\\hbox{\\b{o}}}" "\\accent'27"]
		set texoth4  [list     "\\copyright$seplater" "\\ddag$seplater"]
		    
		set texall   [concat  $texas $texcas $texes $texces $texis $texcis $texos $texcos  $texus $texcus $texlig $texoth1 $texcoth1 $texoth2 $texoth3 $texoth4]
	}

	      
	############################################################
	#														                               #
	#				 List of regular expressions		                   #
	#			     For 	the	reg	exp	is				                   #
	#			      \`[ \t]*(a|{[	\t]*a[ \t]*})		                 #
	#                 \c c needs the space but \c{c} does not  #
	#                 \i may have trailing spaces              #
	#														                               #
	############################################################

	if {$ww == "0"} {      
		set ws {[ \t]*}
		set sp {[ \t]}
		set sep { *( |\b)}
		set a [rexp a]
		set regas  [list  "\\\\`$a" "\\\\'$a" "\\\\\\^$a"  "\\\\~$a"  "\\\\\\$quote$a" "\\\\aa$sep"]
		set a [rexp A]
		set regcas [list  "\\\\`$a" "\\\\'$a" "\\\\\\^$a"  "\\\\~$a"  "\\\\\\$quote$a" "\\\\AA$sep"]
		set e [rexp e]
		set reges  [list  "\\\\`$e"  "\\\\'$e" "\\\\\\^$e"  "\\\\\\$quote$e"]
		set e [rexp E]
		set regces [list  "\\\\`$e"  "\\\\'$e" "\\\\\\^$e"  "\\\\\\$quote$e"]
		set i [rexp \\i$ws]
		set regis  [list  "\\\\`$i" "\\\\'$i"  "\\\\\\^$i"  "\\\\$quote$i"]
		set i [rexp I]
		set regcis  [list "\\\\`$i" "\\\\'$i"  "\\\\\\^$i"  "\\\\$quote$i"]
		set o [rexp o]
		set regos   [list "\\\\`$o" "\\\\'$o"  "\\\\\\^$o"  "\\\\~$o"  "\\\\$quote$o"]
		set o [rexp O]
		set regcos   [list "\\\\`$o" "\\\\'$o"  "\\\\\\^$o"  "\\\\~$o"  "\\\\$quote$o"]
		set u [rexp u]
		set regus    [list "\\\\`$u" "\\\\'$u"  "\\\\\\^$u"  "\\\\$quote$u"]
		set u [rexp U]
		set regcus   [list "\\\\`$u" "\\\\'$u"  "\\\\\\^$u"  "\\\\$quote$u"]
		set reglig   [list "\\\\oe$sep"  "\\\\OE$sep" "\\\\ae$sep" "\\\\AE$sep"]
		set regoth1  [list "\\\\c$sp[rexp c]|\\\\c{$ws\(c\)$ws}"  "\\\\~[rexp n]"  "\\\\\\$quote[rexp y]"  "\\\\dots$sep"]
		set regcoth1 [list "\\\\c$sp[rexp C]|\\\\c{$ws\(C\)$ws}"  "\\\\~[rexp N]"  "\\\\\\$quote[rexp Y]"]
		set regoth2  [list "\\\\ss$sep"  "\\?`" "\\!`" "\\\\pounds$sep" "\\\\o$sep" "\\\\O$sep" "\\\\S$sep" "\\\\P$sep"]
		set regoth3  [list "{\\\\leavevmode\\\\raise.585ex\\\\hbox{\\\\b{a}}}" "{\\\\leavevmode\\\\raise.6ex\\\\hbox{\\\\b{o}}}" "\\\\accent'27"]
		set regoth4  [list     "\\\\copyright$sep" "\\\\ddag$sep"]
		
		set regall [concat $regas $regcas $reges $regces $regis $regcis $regos $regcos $regus $regcus $reglig $regoth1 $regcoth1 $regoth2 $regoth3 $regoth4]
	}

	############################################################
	set mark {\\\\my-mark-fmd//}
	set space {[ ]+}
	set len [llength $all]
	set i 0
	while {$i < $len} {
		set c [lindex $all $i]
		if {$ww == "1"} {set s [lindex $texall $i]
			regsub -all "$c" $text "$s" text 
		} else {
			set s [lindex $regall $i]
			regsub -all "$s" $text "$mark$c$mark" text
		}
		incr i
	}
	if {$ww == "1"} {   
		regsub -all "$seplater$space" $text "\\\\ " text  
		regsub -all "$seplater\\\\" $text "\\\\" text 
		regsub -all "\($seplater\)\(\\W\)" $text "\\2" text
		regsub -all "$seplater" $text " " text
	} else {
		regsub -all "$mark\\\\ " $text " " text
		regsub -all "$mark" $text "" text
	}
	
	replaceText $position $endselection $text	
}
