## -*-Tcl-*- (install)
 # ###################################################################
 #  AEVT - Parsing functions for AEGizmo strings
 # 
 #  FILE: "aevt.tcl"
 #                                    created: 26/7/97 {6:44:05 pm} 
 #                                last update: 20/12/97 {6:50:48 pm} 
 #                                    version: 1.0.1
 #  Author: Jonathan Guyer
 #  E-mail: <j-guyer@nwu.edu>
 #    mail: Northwestern University
 #          Evanston, IL
 #  
 # ###################################################################
 ##

alpha::extension aevt 1.0.1 {
	# turns out we don't need this, but it might come in handy
	
	# global variable for the Startup Disk 
	# (as sometimes returned by the Finder)
	set STARTUP [aevt::TEXT [aevt::keywordValue ---- \
	  [lindex [aevt::event [objectProperty 'MACS' pnam \
	    "obj \{want:type(prop), from:'null'(), \
		form:prop, seld:type(sdsk)\}"]] 2]]]	
}

namespace eval aevt {}

## 
 # Copyright (c) 1997  Jonathan	Guyer
 # 
 # This	program	is free	software; you can redistribute it and/or modify
 # it under	the	terms of the GNU General Public	License	as published by
 # the Free	Software Foundation; either	version	2 of the License, or
 # (at your	option)	any	later version.
 # 
 # This	program	is distributed in the hope that	it will	be useful,
 # but WITHOUT ANY WARRANTY; without even the implied warranty of
 # MERCHANTABILITY or FITNESS FOR A	PARTICULAR PURPOSE.	 See the
 # GNU General Public License for more details.
 # 
 # You should have received	a copy of the GNU General Public License
 # along with this program;	if not,	write to the Free Software
 # Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,	USA.
 ##

## 
 # This	is a set of	AEGizmo	parsing	routines, reasonably compliant with	Jens
 # Alfke's "formidable-looking BNF grammar specification".	I'd	like to	hear
 # about anything it doesn't parse correctly.  Given an	'aevt\ansr{...}', a
 # 'GURL\GURL{...}', or	some subset	of an AEGizmo string, you can parse	to a
 # list	of lists; keyword searches are then	easy, i.e,
 # 
 # aevt::event "aevt\\ansr\{'----':Yabba dabba	doo!\}"
 # 
 #	   returns
 # 
 # aevt	ansr {{----	{Yabba	dabba doo!}}}
 # 
 #	   and
 # 
 # aevt::TEXT [	\
 #	   aevt::keywordValue "----" \
 #	   [lindex [aevt::event	"aevt\\ansr\{'----':Yabba dabba doo!\}"] 2]]
 # 
 #	   returns
 # 
 # Yabba dabba doo!
 # 
 # 
 # I decided to	write this up because I	got	tired of always	having to check
 # for special cases in	my regexp's. I hope others find it useful too.
 ##

## 
 # ident ::= identchar (identchar |	digit)*	   Padded/truncated
 #			 ' character* '					   to exactly 4	chars
 ##
proc aevt::ident {chars {integerOK 0}} {
	set identchar	{[^][(){} \r\t\n0-9':,@]}
	set chars [string trimleft $chars]
	if {[regexp -indices "^${identchar}(${identchar}|\[0-9\])*" $chars pos]} {
		set ident [format "%-4s" [string range $chars 0 [lindex $pos 1]]]
		set chars [string range $chars [expr [lindex $pos 1] + 1] end]
		set result [list "$ident" "$chars"]
	} elseif {[regexp -indices "^'(\[^'\]*)'" $chars blah pos]} {
		set ident [format "%-4s" [string range $chars [lindex $pos 0] [lindex $pos 1]]]
		set chars [string range $chars [expr [lindex $pos 1] + 2] end]
		set result [list "$ident" "$chars"]
	} else {
		error "no ident"
	}
	return $result
}

## 
 # data	::=	@		   Gets appropriate data from fn param
 #			integer	   'shor' or 'long' unless	coerced
 #			ident	   A 4-char type code ('type')	unless coerced
 #			string	   Unterminated text; 'TEXT' type unless coerced
 #			hexstring  Raw	hex	data; must be coerced to some type!
 ##
proc aevt::data {chars} {
	set chars [string trimleft $chars]
	if {[string index $chars 0] == "@"} {
		set data "@"
		set type "@"
		set chars [string range $chars 1 end]
	} elseif {[regexp -indices {^-?[0-9]+} $chars pos]} {
		set pos [lindex $pos 1]
		set data [string range $chars 0 $pos]
		set chars [string range $chars [incr pos] end]
		set type "integer"
	} elseif {![catch {set data [aevt::ident $chars]} errMsg]} {
		set chars [lindex $data 1]
		set data [lindex $data 0]
		set type "ident"
	} elseif {$errMsg != "no ident"} {
		aevt::rethrow $errMsg		
	} elseif {[regexp -indices {^[^]*} $chars pos]} {
		set pos [lindex $pos 1]
		set data [string range $chars 0 $pos]
		set chars [string range $chars [incr pos] end]
		set type "string"
	} elseif {[regexp -indices {^[0-9a-fA-F \r\t\n]*} $chars pos]} {
		set pos [lindex $pos 1]
		set data [string range $chars 0 $pos]
		set chars [string range $chars [incr pos] end]
		set type "hexstring"
	} else {
		error "no data"
	}
	return [list $data $chars $type]
}

## 
 # objectlist ::= blank			  Comma-separated list	of things
 #				  obj [	, obj ]*
 ##
proc aevt::objectlist {chars} {
	set chars [string trimleft $chars]
	set result ""
	if {[string index $chars 0] == "\["} {
		set chars [string trimleft [string range $chars 1 end]]
		if {[string index $chars 0] != "\]"} {
			while 1 {
				set obj [aevt::obj $chars]
				lappend result [lindex $obj 0]
				set chars [string trim [lindex $obj 1]]
				set next [string index $chars 0]
				set chars [string range $chars 1 end]
				if {$next == "\]"} {
					break
				} elseif {$next != ","} {
					error "aeBuildSyntaxNoCloseBracket" "" 14
				}		
			}
		}
		set result [list $result $chars]
	} else {
		error "no objectlist"
	}
	return $result
}

## 
 # keywordpair ::= ident : obj		  Keyword/value pair
 ##
proc aevt::keywordpair {chars} {
	set chars [string trim $chars]
	if {![catch {set ident [aevt::ident $chars]} errMsg]} {
		set keyword [lindex $ident 0]
		set chars [string trim [lindex $ident 1]]
		if {[string index $chars 0] == ":"} {
			set chars [string range $chars 1 end]
			set value [aevt::obj $chars]
			set result [list [list $keyword [lindex $value 0]] [lindex $value 1]]
		} else {
			error "aeBuildSyntaxNoColon" "" 17
		}
	} elseif {$errMsg == "no ident"} {
		error "aeBuildSyntaxNoKey" "" 16
	} else {
		aevt::rethrow $errMsg
	}
	return $result
}

## 
 # keywordlist ::= blank				List of said pairs
 #				   keywordpair [ , keywordpair ]*
 ##
proc aevt::keywordlist {chars} {
	set chars [string trim $chars]
	set result ""
	if {[string index $chars 0] == "\{"} {
		set chars [string trimleft [string range $chars 1 end]]
		if {[string index $chars 0] != "\}"} {
			while 1 {
				set pair [aevt::keywordpair $chars]
				lappend result [lindex $pair 0]
				set chars [string trimleft [lindex $pair 1]]
				set next [string index $chars 0]
				set chars [string range $chars 1 end]
				if {$next == "\}"} {
					break
				} elseif {$next != ","} {
					error "aeBuildSyntaxNoCloseBrace" "" 15
				}
			}
		}
		set result [list "$result" "$chars"]
	} else {
		error "no keywordlist"
	}
	return $result
}

## 
 # structure ::= ( data	)		   Single AEDesc
 #				 [ objectlist ]	   AEList type
 #				 { keywordlist }   AERecord type
 ##
proc aevt::structure {chars} {
	set chars [string trimleft $chars]
	set result ""
	set errMsg ""
	if {[string index $chars 0] == "("} {
		if {![catch {set data [aevt::data [set temp [string range $chars 1 end]]]} errMsg]} {
			if {[string index [set temp [lindex $data 1]] 0] == ")"} {
				set chars [string range $temp 1 end]
				set data [lindex $data 0]
				set result [list "$data" "$chars"]
			} else {
				error "aeBuildSyntaxNoCloseParen" "" 13
			}
		} elseif {$errMsg != "no data"} {
			aevt::rethrow $errMsg
		} else {
			set temp [string trimleft $temp]
			if {[string index $temp 0] == ")"} {
				set data ""
				set chars [string range $temp 1 end]
				set result [list "$data" "$chars"]
			} else {
				error "aeBuildSyntaxNoCloseParen" "" 13
			}
		}
	} elseif {[catch {set result [aevt::objectlist $chars]} errMsg]} {
		if {$errMsg != "no objectlist"} {
			aevt::rethrow $errMsg
		} elseif {[catch {set result [aevt::keywordlist $chars]} errMsg]} {
			if {$errMsg != "no keywordlist"} {
				aevt::rethrow $errMsg
			} else {
				error "no structure"
			}
		}
	}
	return $result
}

## 
 # obj ::= data				 Single AEDesc; shortcut for (data)
 #		   structure		 Un-coerced structure
 #		   ident structure	 Coerced to some other	type
 ##
proc aevt::obj {chars} {
	set chars [string trim $chars]
	set result ""

	if {![catch {set data [aevt::data $chars]} errMsg]} {
		set obj [lindex $data 0]
		set chars [lindex $data 1]
		if {[lindex $data 2] == "ident"} {
			if {![catch {set result [aevt::structure $chars]} errMsg]} {
				set result [list [list "$obj" "[lindex $result 0]"] \
								"[lindex $result 1]"]
			} elseif {$errMsg == "no structure"} {
				set result [list "$obj" "$chars"]
			} else {
				aevt::rethrow $errMsg
			}
		} else {
			set result [list "$obj" "$chars"]
		}
	} elseif {$errMsg != "no data"} {
		aevt::rethrow $errMsg
	} else {
		set result [aevt::structure $chars]
	}
	return $result
}

## 
 # Return the value	associated with	$keyword in	the	parsed list	$keywordpairs
 # If $reportError,	throw an error,	otherwise return null
 ##
proc aevt::keywordValue {keyword keywordpairs {reportError 1}} {
	
	# Strip user supplied '' quotes, if any
	regexp "^'(.*)'$" $keyword blah keyword
	set keyword [format "%-4s" [string range $keyword 0 3]]
	
	foreach keywordpair $keywordpairs {
		if {[lindex $keywordpair 0] == $keyword} {
			return [lindex $keywordpair 1]
		} 
	}
	if {$reportError} {
		set errs [aevt::TEXT [aevt::keywordValue "errs" $keywordpairs 0]]
		set errn [aevt::keywordValue "errn" $keywordpairs 0]
		if {[string length $errs] || [string length $errn]} {
			error $errs "" $errn 		
		}
	} 
}

## 
 # TEXT	::=	TEXT(text)
 #			text
 ##
proc aevt::TEXT {value} {
	if {[lindex $value 0] == "TEXT"} {
		return [lindex $value 1]
	} elseif {[regexp {([^]*)} $value blah body]} {
		return $body
	} else {
		error "$value is not valid TEXT"
	}
}

## 
 # "aevt::toTEXT" --
 # 
 #  Coerce list to an AE list of TEXT items
 ##
proc aevt::toTEXT {list} {
	set TEXT ""
	foreach item $list {
		lappend TEXT "$item"
	}
	set TEXT [join $TEXT ", "]
	if {[llength $list] > 1} {
		set TEXT "\[$TEXT\]"
	}
	return $TEXT
}



## 
 # event ::= ident '\' ident keywordlist
 ##
proc aevt::event {chars} {
	if {[set pos [string first "\\" $chars]] >= 0} {
		set class [string range $chars 0 [expr $pos - 1]]
		set chars [string range $chars [expr $pos + 1] end]
		
		# Make sure $class is formatted correctly
		set class [aevt::ident $class]
		set class [lindex $class 0]
		
		set event [aevt::ident $chars]
		set chars [lindex $event 1]
		set event [lindex $event 0]
		
		set parameters [aevt::structure $chars]
		set parameters [lindex $parameters 0]
		
		return [list $class $event $parameters]
	}
	error "$chars is not a valid AE event"
}

## 
 # Call	with $msg in the event you 'catch' an error	you	don't want.
 # All this	saves is having	to declare 'global errorCode' all over your	code
 ##
proc aevt::rethrow {msg} {
	global errorCode
	
	error $msg "" $errorCode
}

