#
# strings.tcl (Mark Nagata and Tom Scavo and Vince Darley)
#

namespace eval quote {}
namespace eval text {}
## 
 # -------------------------------------------------------------------------
 # 
 # "quote::" --
 # 
 # Manipulate string so search and insertion procedures work as expected.
 # These files have been both renamed and rewritten from the former
 # 'quoteExpr' procs.  They fix a number of bugs, and make their purpose
 # clear.  There were numerous examples throughout Alpha's Tcl code which
 # used the wrong quote function under the old scheme.
 # 
 # quote::Find
 # 
 # 	use this for 'glob' type searches.
 # 	
 # quote::Regfind
 # 
 #  use this for regexp searches
 #  
 # quote::Insert
 # 
 #  Quotes any block of text captured from a window so it can be used as a 
 #  Tcl string. e.g. 'set a [quote::Insert [getSelect]] ; eval insertText $a'
 #  will work correctly.  Can be used to generate procedures on the fly,
 #  especially to add to your prefs.tcl:
 #   set a [quote::Insert [getSelect]]
 #   addUserLine "proc foo \{\} \{ return \"$a\" \}"
 # 
 # quote::Regsub
 # 
 #  use this for the replacement expression.  A common usage might look
 #  like this:
 #   
 #   regsub -all [quote::Regfind $from] [read $cid] [quote::Regsub $to] out
 # -------------------------------------------------------------------------
 ##
proc quote::Find  str {
	regsub -all {[][\|*+()]} $str {\\&} str
	return $str
}

proc quote::Regfind str {
	regsub -all {[][\$?^|*+()\.\{\}]} $str {\\&} str
	return $str
}

proc quote::Insert str {
	regsub -all {[][\$"\{\}]} $str {\\&} str
	regsub -all "\[\r\n\]" $str "\\r" str
	regsub -all "\t" $str "\\t" str
	return $str
}

proc quote::Regsub str {
	regsub -all {(\\|&)} $str {\\&} str
	return $str
}

## 
 # -------------------------------------------------------------------------
 # 
 # "quote::Prettify" --
 # 
 #  Since we're supposed to be a LaTeX editor, we handle symbols with
 #  TeX in a bit differently
 # -------------------------------------------------------------------------
 ##
proc quote::Prettify str {
	set a [string toupper [string index $str 0]]
	regsub -all {([^A-Z])([A-Z])} [string range $str 1 end] {\1 \2} b
	regsub -all {((La|Bib|Oz) )?Te X} $a$b {\2TeX } a
	regsub -all {::} $a {-} a
	return $a
}
proc quote::Menuify str {
	set a [string toupper [string index $str 0]]
	regsub -all {([^A-Z])([A-Z])} [string range $str 1 end] {\1 \2} b
	append a $b
}
## 
 # -------------------------------------------------------------------------
 # 
 # "quote::WhitespaceReg" --
 # 
 #  Quote a string so you can search for it ignoring all problems with
 #  whitespace: all sequences of space/tab/cr are treated alike.
 # -------------------------------------------------------------------------
 ##
proc quote::WhitespaceReg { str } { 
	regsub -all "\[ \t\r\]+" $str {[ \t\r]+} str
	return $str
}

## 
 # -------------------------------------------------------------------------
 # 
 # "lremove" --
 # 
 #  removes items from a list
 #  
 #  options are '-all' to remove all, and -glob, -exact or -regexp
 #  for search type.  '-exact' is the default. '--' terminates options.
 #  
 #  lremove ?-opts? l args
 #  
 #  Note: if you want to remove all items of list 'b' from list 'a',
 #  the following is incorrect: lremove $a $b, you must use
 #  'eval lremove [list $a] $b', so that b is expanded.
 #  
 #  There is now a new option -l which treats the extra args as lists,
 #  so you can do lremove -l $a $b if you want.
 # -------------------------------------------------------------------------
 ##
proc lremove {args} {
    set opts(-all) 0
	set type "-exact"
    getOpts
    set l [lindex $args 0]
	if [info exists opts(-glob)] { set type "-glob" }
	if [info exists opts(-regexp)] { set type "-regexp" }
	if [info exists opts(-l)] { 
		set args [join [lreplace $args 0 0] " "]
	} else {
		set args [lreplace $args 0 0]
	}
	foreach i $args {
		if {[set ix [lsearch $type $l $i]] == -1} continue
		set l [lreplace $l $ix $ix]
		if {$opts(-all)} {
			while {[set ix [lsearch $type $l $i]] != -1} {
				set l [lreplace $l $ix $ix]
			}
		}
	}
    return $l
}

## 
 # -------------------------------------------------------------------------
 # 
 # "getOpts" --
 # 
 #  Rudimentary option passing.  Uses upvar to get to the 'args' list of
 #  the calling procedure and scans that.  Option information is stored
 #  in the 'opts' array of the calling procedure.
 #  
 #  Options are assumed to be flags, unless they occur in the optional
 #  parameter list.  Then they are variables which take a value; the
 #  next item in the args list.  If an item is a pair, then the first
 #  is the var name and the second the number of arguments to give it.
 # -------------------------------------------------------------------------
 ##
proc getOpts {{take_value ""} {set "set"}} {
	upvar args a
	upvar opts o
	while {[string match \-* [set arg [lindex $a 0]]]} {
		set a [lreplace $a 0 0]
		if {$arg == "--"} {
			return
		} else {
			if {[set idx [lsearch -regexp $take_value \
			  "^-?[string range $arg 1 end]( .*)?$"]] == -1} {
				set o($arg) 1
			} else {
				if {[llength [set the_arg [lindex $take_value $idx]]] == 1} {
					$set o($arg) [lindex $a 0]
					set a [lreplace $a 0 0]
				} else {
					set numargs [expr [lindex $the_arg 1] -1]
					$set o($arg) [lrange $a 0 $numargs]
					set a [lreplace $a 0 $numargs]
				}
			}
		}
	}
}

## 
 # -------------------------------------------------------------------------
 # 
 # "ensureset" --
 # 
 #  Ensure the given variable is set, if it is unset, set it to the given
 #  value.  This works with both variables and array elements, including
 #  things which contain spaces etc.
 # -------------------------------------------------------------------------
 ##
proc ensureset {v {val ""}} {
	if [uplevel [list info exists $v]] { return [uplevel [list set $v]] }
	return [uplevel [list set $v $val]]
}
## 
 # -------------------------------------------------------------------------
 # 
 # "lunion" --
 # 
 #  Basic use: make sure a given list variable contains each element 
 #  of 'args'
 #  
 #  "llunion" --
 #  
 #  Advanced use: make sure a given list variable and index contains
 #  an element whose i'th index matches the i'th index of one of 'args'.
 #  In this case we call the proc with a list {var i} as first argument.
 # -------------------------------------------------------------------------
 ##
proc lunion {var args} {
	upvar $var a
	if ![info exists a] {
		set a $args
		return
	} else {
		foreach item $args {
			if {[lsearch $a $item] == -1} {
				lappend a $item
			}
		}
	}
}
	
proc llunion {var idx args} {
	upvar $var a
	if ![info exists a] {
		set a $args
		return
	} else {
		foreach item $args {
			set add 1
			foreach i $a {
				if {[lindex $i $idx] == [lindex $item $idx]} {
					set add 0
					break
				}
			}
			if {$add} {
				lappend a $item
			}
		}
	}
}

proc lunique {l} {
	set lout ""
	foreach f $l {
		if {![info exists silly($f)]} {
			set silly($f) 1
			lappend lout $f
		}
	}
	return $lout
}
			
proc lreverse {l} {
	if {[llength $l] > 1} {
		set first [lindex $l 0]
		set l [lreverse [lrange $l 1 end]]
		lappend l $first
	}
	return $l
}

proc lcontains {l e} {
	upvar $l ll
	if {[info exists ll] && [lsearch -exact $ll $e] != -1} {
		return 1
	} else {
		return 0
	}
}

## 
 # -------------------------------------------------------------------------
 # 
 # "llindex" --
 # 
 #  Find the first index of a given list within another list.  
 # -------------------------------------------------------------------------
 ##
proc llindex {l e args} {
	upvar $l ll
	if ![info exists ll] { return -1 }
	if {$args == ""} {
		return [lsearch -exact $ll $e]
	} else {
		set i 0
		set len [llength $args]
		while {$i < [llength $ll] - $len} {
			if {[lindex $ll $i] == $e} {
				set range [lrange $ll [expr $i +1] [expr $i + $len]]
				for {set j 0} {$j < $len} {incr j} {
					if {[lindex $args $j] != [lindex $range $j]} {
						break
					}
				}
				if {$j == $len} { return $i}
			}
			incr i
		}
		return -1
	}
}

########################################
#									   #
#	A few random lisp'ish functions.   #
#									   #
########################################

proc car {l} {lindex $l 0}
proc cadr {l} {lindex $l 1}
proc caddr {l} {lindex $l 2}
proc cadddr {l} {lindex $l 3}
proc caddddr {l} {lindex $l 4}
proc cdr {l} {lrange $l 1 end}
proc cddr {l} {lrange $l 2 end}
proc cons {e l} {concat [list $e] $l}
proc mapcar args {return [eval map $args]}

proc map {func l} {
	set out {}
	foreach el $l {
		lappend out [eval $func [list $el]]
	}
	return $out
}

# Returns a modified text string if the string $text is non-null, 
# and the null string otherwise.  The argument 'operation' is a 
# string directing 'doSuffixText' to either "insert" or "remove" 
# $suffixString to/from each line of $text.
proc doSuffixText {operation suffixString text} {
	if {$text == ""} {return ""}
	set suff [quote::Find $suffixString]
	if {$operation == "insert"} {
		set str ${suffixString}\r
		regsub -all \r $text $str text
	} elseif {$operation == "remove"} {
		set str ${suff}\r
		regsub -all $str $text \r text
	}
	return $text
}

# Returns a modified text string if the string $text is non-null, 
# and the null string otherwise.  The argument 'operation' is a 
# string directing 'doPrefixText' to either "insert" or "remove" 
# $prefixString to/from each line of $text.  See latexEngine.tcl
# for an example.
proc doPrefixText {operation prefixString text} {
	if {$text == ""} {return ""}
	set pref [quote::Find $prefixString]
	if {$operation == "insert"} {
		set trailChar ""
		set textLen [string length $text]
		if {[string index $text [expr $textLen-1]] == "\r"} {
			set text [string range $text 0 [expr $textLen-2]]
			set trailChar "\r"
		}
		set str \r$prefixString
		regsub -all \r $text $str text
		return $prefixString$text$trailChar
	} elseif {$operation == "remove"} {
		regsub -all \r$pref $text \r text
		regsub ^$pref $text "" text
		return $text
	}
}

proc text::british {v} {
	uplevel "regsub -all -nocase {(Colo)r} \[set $v\] {\\1ur} $v"
}

rename getAscii {}
proc getAscii {} {
	set c [lookAt [getPos]]
	scan $c %c decVal
	set asOctal [format %o $decVal]
	set asHex   [format %x $decVal]
	alertnote "saw a \"$c\", $decVal -decimal,\
	  \\$asOctal -octal, x$asHex -hex"
}

# nabbed from html mode
set text::_Ascii "\001\002\003\004\005\006\007\010\011\012\013\014\015\016\017"
append text::_Ascii "\020\021\022\023\024\025\026\027\030\031\032\033\034\035\036\037"
append text::_Ascii " !\"#\$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ"
append text::_Ascii "\[\\\]^_`abcdefghijklmnopqrstuvwxyz{|}~\177"
append text::_Ascii ""
append text::_Ascii ""
proc text::Ascii {char {num 0}} {
	if {$char == ""} {return 0}
	global text::_Ascii
	if {$num} {
		if {$char > 256 || $char < 1} { beep ; message "text::Ascii called with bad argument" }
		return [string index ${text::_Ascii} [expr $char - 1]]
	} else {
		return [expr 1 + [string first $char ${text::_Ascii}]]
	}
}

# Useful for -command flag of 'lsort'.
proc sortByTail {one two} {
	string compare [file tail $one] [file tail $two]
}

